# -*- coding: utf-8 -*-

import pytest

from simmate.conftest import copy_test_files
from simmate.workflows.utilities import (
    get_workflow_types,
    get_list_of_all_workflows,
    get_list_of_workflows_by_type,
    get_workflow,
    load_results_from_directories,
    get_unique_parameters,
)


def test_get_workflow_types():
    assert get_workflow_types() == [
        "customized",
        "diffusion",
        "dynamics",
        "electronic-structure",
        "population-analysis",
        "relaxation",
        "restart",
        "static-energy",
    ]


def test_list_of_all_workflows():

    assert get_list_of_all_workflows() == [
        "customized.vasp.user-config",
        "diffusion.vasp.neb-all-paths",
        "diffusion.vasp.neb-from-endpoints",
        "diffusion.vasp.neb-from-images",
        "diffusion.vasp.neb-single-path",
        "dynamics.vasp.mit",
        "electronic-structure.vasp.matproj-full",
        "population-analysis.vasp.badelf-matproj",
        "population-analysis.vasp.bader-matproj",
        "population-analysis.vasp.elf-matproj",
        "relaxation.vasp.matproj",
        "relaxation.vasp.mit",
        "relaxation.vasp.neb-endpoint",
        "relaxation.vasp.quality00",
        "relaxation.vasp.quality01",
        "relaxation.vasp.quality02",
        "relaxation.vasp.quality03",
        "relaxation.vasp.quality04",
        "relaxation.vasp.staged",
        "restart.simmate.automatic",
        "static-energy.vasp.matproj",
        "static-energy.vasp.mit",
        "static-energy.vasp.neb-endpoint",
        "static-energy.vasp.quality04",
    ]


def test_list_of_workflows_by_type():

    assert get_list_of_workflows_by_type("static-energy") == [
        "static-energy.vasp.matproj",
        "static-energy.vasp.mit",
        "static-energy.vasp.neb-endpoint",
        "static-energy.vasp.quality04",
    ]

    with pytest.raises(TypeError):
        get_list_of_workflows_by_type("non-existant-type")


def test_get_workflow():

    from simmate.workflows.static_energy import (
        StaticEnergy__Vasp__Matproj as workflow,
    )

    assert get_workflow("static-energy.vasp.matproj") == workflow


def test_get_unique_paramters():
    assert get_unique_parameters() == [
        "command",
        "copy_previous_directory",
        "diffusion_analysis_id",
        "directory",
        "directory_new",
        "directory_old",
        "input_parameters",
        "is_restart",
        "migrating_specie",
        "migration_hop",
        "migration_hop_id",
        "migration_images",
        "nsteps",
        "pre_sanitize_structure",
        "pre_standardize_structure",
        "source",
        "structure",
        "supercell_end",
        "supercell_start",
        "temperature_end",
        "temperature_start",
        "time_step",
        "updated_settings",
        "workflow_base",
    ]


@pytest.mark.django_db
def test_load_results_from_directories(tmpdir):

    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="",  # copy over the entire folder
    )

    load_results_from_directories(base_directory=tmpdir)



# -*- coding: utf-8 -*-

from simmate.workflow_engine import Workflow
from simmate.calculators.vasp.tasks.population_analysis import (
    MatprojELF as MPPreBaderTask,
)
from simmate.calculators.vasp.database.population_analysis import (
    MatprojELF as MPELFResults,
)


class PopulationAnalysis__Vasp__ElfMatproj(Workflow):
    s3task = MPPreBaderTask
    database_table = MPELFResults
    description_doc_short = "uses Materials Project settings"



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class RotationMatrix(ErrorHandler):
    """
    This a simple error handler that is active when VASP struggles to find the
    rotation matrix. VASP gives us the suggested fix directly, which is to
    simply increase the symmetry precision.
    """

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["rotation matrix was not found (increase SYMPREC)"]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # increase the precision
        incar["SYMPREC"] = 1e-8

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return "switched SYMPREC to 1e-8"



# -*- coding: utf-8 -*-

from numpy.random import choice

from simmate.toolkit.creators.structure.base import StructureCreator


class PyXtalStructure(StructureCreator):

    # see docs: https://pyxtal.readthedocs.io/en/latest/index.html

    def __init__(
        self,
        composition,  # must be a pymatgen Composition object,
        volume_factor=1.0,
        default_lattice=None,
        tolerance_matrix=None,  # options unique to PyXtal's random_crystal
        spacegroup_include=range(1, 231),
        spacegroup_exclude=[],
    ):

        #!!! this is inside the init because not all users will have this installed!
        try:
            from pyxtal import pyxtal
        except ModuleNotFoundError:
            #!!! I should raise an error in the future
            raise Exception("You must have PyXtal installed to use PyXtalStructure!!")

        # save the module for reference below
        self.pyxtal = pyxtal()

        # save the composition information in the format that PyXtal wants ['Ca', 'N']
        self.species = [element.symbol for element in composition.elements]
        # save the number of Ions in the format that PyXtal wants [3,2]
        self.numIons = [
            int(composition[element.symbol]) for element in composition.elements
        ]
        # save the volume factor to be used later
        self.volume_factor = volume_factor
        # if the user supplied a lattice, it should be a 3x3 matrix -- NOT a pymatgen lattice object
        #!!! should I add a check that this is a matrix here?
        self.default_lattice = default_lattice
        #!!! TO-DO: I need to allow the user to specify the tolerance matrix
        if tolerance_matrix:
            print(
                "Simmate's wrapper for PyXtal does not currently support setting "
                "the tolerance matrix for PyXtal's random_crystal(). If you want this "
                "option added, please contact the Simmate developers."
            )

        # make a list of spacegroups that we are allowed to choose from
        self.spacegroup_options = [
            sg for sg in spacegroup_include if sg not in spacegroup_exclude
        ]
        # Note, there's no easy way to see which spacegroups are compatible with
        # our composition, so this list is updated as attempts fail in the
        # create_structure method.

    def create_structure(self, spacegroup=None):

        # If a spacegroup is not specified, grab a random one from our options.
        # The requested_spacegroup let's us know below if the method wanted a
        # given spacegroup -- which tells us whether or not we should raise an
        # error if the spacegroup is incompatible with our composition
        if not spacegroup:
            # randomly select a symmetry system
            spacegroup = choice(self.spacegroup_options)
            requested_spacegroup = False
        else:
            requested_spacegroup = True

        from pyxtal.msg import Comp_CompatibilityError

        try:
            # now make the new structure using PyXtal's crystal.random_crystal function
            # NOTE: all of these options are set in the init except for the spacegroup
            # No structure is returned. Instead they just attach it to the pyxtal object
            self.pyxtal.from_random(
                group=int(
                    spacegroup
                ),  # !!! why do I need int() here? This should already work without it...
                species=self.species,
                numIons=self.numIons,
                factor=self.volume_factor,
                lattice=self.default_lattice,
            )
            # tm=<pyxtal.crystal.Tol_matrix object> # not supported right now!
        except Comp_CompatibilityError as exception:
            # if the user wanted this spacegroup, raise the error
            if requested_spacegroup:
                raise exception
            # otherwise remove the spacegroup from our allowed list
            self.spacegroup_options.remove(spacegroup)
            # And retry! Note: recursion is a shortcut here, but shouldn't be an issue
            self.create_structure()

        # note that we have a PyXtal structure object, not a PyMatGen one!
        # The PyXtal structure has a *.valid feature that lets you know if structure creation was a success
        # so we will check that
        if self.pyxtal.valid:
            # if it is valid, we want the PyMatGen structure object, which is at *.struct
            return self.pyxtal.to_toolkit()
        else:
            # if it is not valid, the structure creation failed
            return False



"""
Defines all Django settings for Simmate.

Note, some settings are dynamically determined, such as the installed apps and
database backend being used.
"""

import os
import sys
from pathlib import Path

import yaml

# from django.core.management.utils import get_random_secret_key

import dj_database_url  # needed for DigitalOcean database connection

from simmate import website  # needed to specify location of built-in apps
from simmate.utilities import get_directory, get_conda_env

# --------------------------------------------------------------------------------------

# SET MAIN DIRECTORIES

# We check for extra django apps in the user's home directory and in a "hidden"
# folder named "simmate/extra_applications".
# For windows, this would be something like...
#   C:\Users\exampleuser\simmate\extra_applications
# Note, we use get_directory in order to create that folder if it does not exist.
SIMMATE_DIRECTORY = get_directory(os.path.join(Path.home(), "simmate"))

# This directory is where simmate.website is located and helps us indicate
# where things like our templates or static files are located. We find this
# by looking at the import path to see where python installed it.
DJANGO_DIRECTORY = os.path.dirname(os.path.abspath(website.__file__))

# Some settings below also depend on the conda env name. This makes switching
# between different databases and settings as easy as activating different
# conda environments
CONDA_ENV = get_conda_env()

# --------------------------------------------------------------------------------------

# ENVIORNMENT VARIABLES

# There are a number of settings that we let the user configure via enviornment
# variables, which helps control things when we want to launch a website server.
# We check for these variables in the enviornment, and if they are not set,
# they fall back to a default.

# Don't run with debug turned on in production!
# For DigitalOcean, we try grabbing this from an enviornment variable. If that
# variable isn't set, then we assume we are debugging. The == at the end converts
# the string to a boolean for us.
DEBUG = os.getenv("DEBUG", "True") == "True"

# To make this compatible with DigitalOcean, we try to grab the allowed hosts
# from an enviornment variable, which we then split into a list. If this
# enviornment variable isn't set yet, then we just defaul to the localhost.
ALLOWED_HOSTS = os.getenv("DJANGO_ALLOWED_HOSTS", "127.0.0.1,localhost").split(",")

# This decides wether we use a local sqlite database. The order of priority for
# databases goes...
#   (1) Check for a "database.yaml" file in the user's home directory
#   (2) Check for the USE_LOCAL_DATABASE enviornment variable (defaults to True)
#   (3) If USE_LOCAL_DATABASE=False, then check for DATABASE_URL enviornment variable
USE_LOCAL_DATABASE = os.getenv("USE_LOCAL_DATABASE", "True") == "True"
# this is not a typical Django setting

# Keep the secret key used in production secret!
# For DigitalOcean, we grab this secret key from an enviornment variable.
# If this variable isn't set, then we instead generate a random one.
SECRET_KEY = os.getenv(
    "DJANGO_SECRET_KEY", "pocj6cunub4zi31r02vr5*5a2c(+_a0+(zsswa7fmus^o78v)r"
)
# !!! I removed get_random_secret_key() so I don't have to sign out every time
# while testing my server. I may change this back in the future.

# --------------------------------------------------------------------------------------

# DATBASE CONNECTION

# There are three types of database files that we check for -- in order of priority:
#   1. database.yaml
#   2. my_env-database.yaml
#   3. my_env-database.sqlite3 (if USE_LOCAL_DATABASE=True) <-- and create this if doesn't exist
#   4. use a DATABASE_URL env variable
DATABASE_YAML = os.path.join(SIMMATE_DIRECTORY, "database.yaml")
CONDA_DATABASE_YAML = os.path.join(SIMMATE_DIRECTORY, f"{CONDA_ENV}-database.yaml")
CONDA_DATABASE_SQLITE3 = os.path.join(
    SIMMATE_DIRECTORY, f"{CONDA_ENV}-database.sqlite3"
).strip("-")
# if the user is in the (base) env or not using conda, then we will have a
# value of "-database.sqlite3". We remove the starting "-" here.


# Our 1st priority is checking for a "simmate/database.yaml" file
if os.path.exists(DATABASE_YAML):
    with open(DATABASE_YAML) as file:
        DATABASES = yaml.full_load(file)

# Our 2nd priority is checking for a file like "/simmate/my_env-database.yaml
elif os.path.exists(CONDA_DATABASE_YAML):
    with open(CONDA_DATABASE_YAML) as file:
        DATABASES = yaml.full_load(file)

# This is the default behavior
# Our 3rd prioirity is a local sqlite database name "my_env-database.sqlite3"

elif USE_LOCAL_DATABASE is True:
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.sqlite3",
            "NAME": CONDA_DATABASE_SQLITE3,
        }
    }

# Lastly, if we make it to this point, we are likely using DigitalOcean and
# running a server. When DigitalOcean runs the "collectstatic" command, we don't
#  want to connect any database. So we use the "sys" library to look at the
# command and ensure it doesn't involve "collectstatic". Otherwise we use the
# URL that is set with our enviornment variable.
elif len(sys.argv) > 0 and sys.argv[1] != "collectstatic":
    # ensure that we have the database URL properly configured in DigitalOcean
    if os.getenv("DATABASE_URL", None) is None:
        raise Exception("DATABASE_URL environment variable not defined")
    # Now connect our DigitalOcean to this database!
    DATABASES = {
        "default": dj_database_url.parse(os.environ.get("DATABASE_URL")),
    }

# Here is an example of connecting to a Postgres server normally...
# DATABASES = {
#     "default": {
#         "ENGINE": "django.db.backends.postgresql_psycopg2",
#         "NAME": "simmate-database-pool",  # default on DigitalOcean is defaultdb
#         "USER": "doadmin",
#         "PASSWORD": "dibi5n3varep5ad8",
#         "HOST": "db-postgresql-nyc3-09114-do-user-8843535-0.b.db.ondigitalocean.com",
#         "PORT": 25061,
#         "OPTIONS": {"sslmode": "require"},  # !!! is this needed?
#         # "CONN_MAX_AGE": 0,  # set this to higher value for production website server
#     }
# }

# --------------------------------------------------------------------------------------

# INSTALLED APPS

# OPTIMIZE: for now, I just load everything, but I will want to allow users to
# to disable some apps from loading. This will allow faster import/setup times
# when running small scripts. One idea is to have an applications_override.yaml
# where the user specifies only what they want. Alternatively, I can use
# a general DATABASE_ONLY keyword or something similar to limit what's loaded.

# List all applications that Django will initialize with. Write the full python
# path to the app or it's config file. Note that relative python paths work too
# if you are developing a new app. Advanced users can remove unnecessary apps
# if you'd like to speed up django's start-up time.
INSTALLED_APPS = [
    #
    # These are all apps that are built by Simmate
    "simmate.website.core_components.apps.CoreComponentsConfig",
    "simmate.website.third_parties.apps.ThirdPartyConfig",
    "simmate.website.workflows.apps.WorkflowsConfig",
    #
    # These are built-in django apps that we use for extra features
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.sites",
    # Other apps installed with Django to consider
    #   "django.contrib.humanize",
    #   "django.contrib.postgres",
    #   "django.contrib.redirects",
    #   "django.contrib.sitemaps",
    #
    # These are apps created by third-parties that give us extra features
    "crispy_forms",  # django-crispy-forms
    "rest_framework",  # djangorestframework
    "django_filters",  # django-filter
    #
    # Apps for django-allauth that allow sign-on using external accounts
    "allauth",
    "allauth.account",
    "allauth.socialaccount",
    # note: there are extra apps installed based on configuration. See the
    # allauth section at the bottom of this page for more.
    #
    # Other third-party apps/tools to consider. Note that some of these don't
    # need to be installed apps while some also request different setups.
    #   django-ratelimit
    #   dj-stripe
    #   django-unicorn
    #   django-graphene (+ GraphQL)
    #   django-redis
    #
    # These third-party tools are only to help with development and we should
    # try to avoid adding these as installed apps by default.
    #   "django_extensions",  # for development tools
    #   "debug_toolbar",  # django-debug-toolbar  # for debuging and profile-time info
]

# We also check if the user has a "applications.yaml" file. In this file, the
# user can provide extra apps to install for Django. We simply append these
# to our list above
APPLICATIONS_YAML = os.path.join(SIMMATE_DIRECTORY, "applications.yaml")
if os.path.exists(APPLICATIONS_YAML):
    with open(APPLICATIONS_YAML) as file:
        # load the list of extra apps. Since this is really just one line for
        # each, it is loaded as a single string separated by a space.
        extra_apps = yaml.full_load(file).split()
        # now add each app to our list above so Django loads it.
        for app in extra_apps:
            INSTALLED_APPS.append(app)

# --------------------------------------------------------------------------------------

# This sets the default field for primary keys in django models
DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

# "core" here is based on the name of my main django folder
ROOT_URLCONF = "simmate.website.core.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        # I set DIRS below so I can have a single templates folder
        "DIRS": [os.path.join(DJANGO_DIRECTORY, "templates")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

# "core" here is based on the name of my main django folder
WSGI_APPLICATION = "simmate.website.core.wsgi.application"

# Password validation
# https://docs.djangoproject.com/en/3.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": (
            "django.contrib.auth."
            "password_validation.UserAttributeSimilarityValidator"
        ),  # formatted in this odd way because of line length limit for Black
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/3.0/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

DATETIME_INPUT_FORMATS = [
    "%Y-%m-%dT%H:%M",  # this is a custom format I added to get my form widgets working.
    "%Y-%m-%d %H:%M:%S",  # '2006-10-25 14:30:59'
    "%Y-%m-%d %H:%M:%S.%f",  # '2006-10-25 14:30:59.000200'
    "%Y-%m-%d %H:%M",  # '2006-10-25 14:30'
    "%Y-%m-%d",  # '2006-10-25'
    "%m/%d/%Y %H:%M:%S",  # '10/25/2006 14:30:59'
    "%m/%d/%Y %H:%M:%S.%f",  # '10/25/2006 14:30:59.000200'
    "%m/%d/%Y %H:%M",  # '10/25/2006 14:30'
    "%m/%d/%Y",  # '10/25/2006'
    "%m/%d/%y %H:%M:%S",  # '10/25/06 14:30:59'
    "%m/%d/%y %H:%M:%S.%f",  # '10/25/06 14:30:59.000200'
    "%m/%d/%y %H:%M",  # '10/25/06 14:30'
    "%m/%d/%y",  # '10/25/06'
]

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
# !!! Consider removing in the future.
USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.0/howto/static-files/
# collect by running 'python manage.py collectstatic'
STATIC_URL = "/static/"
STATIC_ROOT = os.path.join(DJANGO_DIRECTORY, "static")

# Extra places for collectstatic to find static files.
STATICFILES_DIRS = [os.path.join(DJANGO_DIRECTORY, "static_files")]

# For the dynamically-created structure files, we need to include the static
# directory this to work during local testing. This is NOT allowed in a
# production server, so we don't include it when DEBUG is set to False.
if DEBUG:
    STATICFILES_DIRS += [os.path.join(DJANGO_DIRECTORY, "static")]

# This sets the django-crispy formating style
CRISPY_TEMPLATE_PACK = "bootstrap4"

# Settings for sending emails with my gmail account
# EMAIL_BACKEND = "django.core.mail.backends.smtp.EmailBackend"  # this is the default
EMAIL_HOST = "smtp.gmail.com"
EMAIL_PORT = 587
EMAIL_USE_TLS = True
EMAIL_HOST_USER = "simmate.team@gmail.com"  # os.environ.get('EMAIL_USER')
# !!! REMOVE IN PRODUCTION. Use this instead: os.environ.get('EMAIL_PASSWORD')
EMAIL_HOST_PASSWORD = "example-password-123"

# These settings help configure djangorestframework and our REST API
REST_FRAMEWORK = {
    # The REST framework has both authentication AND permission classes. By
    # default I allow anything, but may revisit these if the user-base ever grows:
    "DEFAULT_AUTHENTICATION_CLASSES": [],
    # Because we have a massive number of results for different endpoints,
    # we want to set results to be paginated by 25 results per page. This
    # way we don't have to set a page limit for every individual endpoint. Note
    # I can consider switching to LimitOffsetPagination in the future, which
    # allows the number of results per page to vary, but I don't do this
    # for now because there's no easy way to set paginator.max_limit in the
    # settings here.
    "DEFAULT_PAGINATION_CLASS": "rest_framework.pagination.PageNumberPagination",
    "PAGE_SIZE": 12,
    # To prevent users from querying too much and bringing down our servers,
    # we set a throttle rate on each user. Here, "anon" represents an anonymous
    # user (not signed-in) while signed-in users have access to higher download
    # rates. Note these are very restrictive because we prefer users to download
    # full databases and use Simmate locally instead.
    "DEFAULT_THROTTLE_CLASSES": [
        "rest_framework.throttling.AnonRateThrottle",
        "rest_framework.throttling.UserRateThrottle",
    ],
    "DEFAULT_THROTTLE_RATES": {"anon": "100/hour", "user": "1000/hr"},
    # We use django-filter to automatically handle filtering from a REST url
    "DEFAULT_FILTER_BACKENDS": [
        "django_filters.rest_framework.DjangoFilterBackend",
        "rest_framework.filters.OrderingFilter",
    ],
    # There are multiple ways to render the data, where we default to a nice HTML
    # view, but also support pure JSON or using Django-REST's interactive API
    "DEFAULT_RENDERER_CLASSES": [
        "rest_framework.renderers.TemplateHTMLRenderer",
        "rest_framework.renderers.BrowsableAPIRenderer",
        "rest_framework.renderers.JSONRenderer",
    ],
}

# Allows the use of iFrames from within Simmate (such as the structure-viewer)
X_FRAME_OPTIONS = "SAMEORIGIN"

# -----------------------------------------------------------------------------

# Extra settings for django-allauth

AUTHENTICATION_BACKENDS = [
    # Needed to login by username in Django admin, regardless of `allauth`
    "django.contrib.auth.backends.ModelBackend",
    # `allauth` specific authentication methods, such as login by e-mail
    "allauth.account.auth_backends.AuthenticationBackend",
]

# simple setting required by allauth. not sure what it does...
SITE_ID = 1

# We start with the providers as an empty dictionary and only fill them
# if client_id and secrets are supplied in as env variables. We do this
# because we don't want broken links when users first start their server.
SOCIALACCOUNT_PROVIDERS = {}
# Other authentication plugins to consider
# "allauth.socialaccount.providers.digitalocean"
# "allauth.socialaccount.providers.orcid"

# Sign-in via Google accounts
GOOGLE_CLIENT_ID = os.getenv("GOOGLE_CLIENT_ID", None)
GOOGLE_SECRET = os.getenv("GOOGLE_SECRET", None)
if GOOGLE_CLIENT_ID and GOOGLE_SECRET:
    INSTALLED_APPS.append("allauth.socialaccount.providers.google")
    SOCIALACCOUNT_PROVIDERS["google"] = {
        "APP": {"client_id": GOOGLE_CLIENT_ID, "secret": GOOGLE_SECRET}
    }

# Sign-in via Github accounts
GITHUB_CLIENT_ID = os.getenv("GITHUB_CLIENT_ID", None)
GITHUB_SECRET = os.getenv("GITHUB_SECRET", None)
if GITHUB_CLIENT_ID and GITHUB_SECRET:
    INSTALLED_APPS.append("allauth.socialaccount.providers.github")
    SOCIALACCOUNT_PROVIDERS["github"] = {
        "APP": {"client_id": GITHUB_CLIENT_ID, "secret": GITHUB_SECRET}
    }

# Initiate social login immediately -- rather than jumping to a separate
# page and then posting.
# SECURITY: consider removing per django-allauth's recommendation
SOCIALACCOUNT_LOGIN_ON_GET = True

# options for login/logoff views
LOGIN_REDIRECT_URL = "/accounts/profile/"  # this is already the default
LOGOUT_REDIRECT_URL = "/accounts/loginstatus/"

# -----------------------------------------------------------------------------

# BUG: To use Django ORM within IPython, Spyder, and Jupyter notebooks, which
# are examples of async consoles, I need to allow unsafe.
# Read more at...
# https://docs.djangoproject.com/en/4.0/topics/async/#async-safety
os.environ["DJANGO_ALLOW_ASYNC_UNSAFE"] = "true"



# -*- coding: utf-8 -*-

import os

# These are dictionaries that tell us which POTCARs we should grab based on
# the type of calculation as well as where to find them
from simmate.calculators.vasp.inputs.potcar_mappings import (
    FOLDER_MAPPINGS,
    PBE_ELEMENT_MAPPINGS,
    PBE_GW_ELEMENT_MAPPINGS,
    # TODO: LDA_ELEMENT_MAPPINGS
)


class Potcar:
    @staticmethod
    def to_file_from_type(
        elements,
        functional,
        filename="POTCAR",
        element_mappings=None,  # actual default is ELEMENT_MAPPINGS
    ):

        # Element objects are passed along with a string representing the
        # desired functional ("PBE", "LDA", or "PBE_GW")
        # The order of the elements list MUST match the POSCAR!

        # If the user wants to override the ELEMENT_MAPPINGS and use different
        # VASP potentials than what we have picked, then they can provide their
        # own dictionary OR pass in an update version of our. For example, they
        # may want to use the PBE potential of "C_h" instead of "C" which uses
        # a harder psuedopotential (useful in high-pressure and molecular
        # calculations). Here, they can do something like...
        #   element_mappings={"C": "C_h", ...} # with all of their other choices
        # or...
        #   element_mappings = PBE_ELEMENT_MAPPINGS.copy()
        #   element_mappings.update({"C": "C_h"})
        # NOTE: remember whereever you use update(), be careful and make sure
        # you update a copy of the imported dictionary and avoid logical bugs.
        # I don't do that here, but I know I'm not mutating the dictionary.
        # Otherwise, if nothing was supplied use our defaults:
        if not element_mappings:
            if functional == "PBE":
                element_mappings = PBE_ELEMENT_MAPPINGS
            if functional == "PBE_GW":
                element_mappings = PBE_GW_ELEMENT_MAPPINGS

        # based on the functional, grab the proper folder location of all POTCARs
        folder_loc = FOLDER_MAPPINGS[functional]

        # we mow need to iterate through each element and find it's POTCAR location
        # we also keep a list of where these are located
        potcar_locations = []
        for element in elements:

            # grab the proper POTCAR symbol based on the functional and element
            potcar_symbol = element_mappings[element.symbol]

            # now let's combine this information for the full path to the POTCAR.
            # The file will be located at /folder_loc/element_symbol/POTCAR
            potcar_loc = os.path.join(folder_loc, potcar_symbol, "POTCAR")

            # add it to our list
            potcar_locations.append(potcar_loc)

        # VASP expect all POTCAR files to be combined into one and in the same
        # order as the POSCAR elements. Let's do that here.
        with open(filename, "w") as combinedfile:
            for potcar in potcar_locations:
                with open(potcar) as singlefile:
                    combinedfile.write(singlefile.read())

    # TODO
    # from_symbol_and_functional --> returns Potential object
    # from_file --> returns Potential object
    # write_from_potential --> takes a Potential object and write file in POTCAR format
    # symbols --> gives list of potcar symbols (Ca, Y_sv, etc.)
    # nelect --> gives total electron count from potcar



# -*- coding: utf-8 -*-

from django import forms

from simmate.utilities import get_chemical_subsystems

# NOTE TO DEV: This form is a special case where we are searching MANY tables
# for information -- so it is therefore not a ModelForm and cannot use
# our mix-ins. I need to think of the best way to leverage these mix-ins without
# creating this form from scratch. For now, this form is effectively a copy/paste
# of `from simmate.website.core_components.filters import Structure`.


class ChemicalSystemForm(forms.Form):

    include_subsystems = forms.BooleanField(
        label="Include Subsytems",
        required=False,
    )
    """
    Whether to include subsystems of the given `chemical_system`. For example,
    the subsystems of Y-C-F would be Y, C, F, Y-C, Y-F, etc..
    """
    # Note: this attribute must be positioned BEFORE chemical_system to ensure
    # it is present in self.cleaned_data before chemical_system is accessed.

    # include_suprasystems = forms.BooleanField(label="Include Subsytems", required=False)
    # TODO: Supra-systems would include all the elements listed AND more. For example,
    # searching Y-C-F would also return Y-C-F-Br, Y-Sc-C-F, etc.

    chemical_system = forms.CharField(
        label="Chemical System",
        max_length=20,
        required=False,
    )
    """
    The chemical system of the structure (e.g. "Y-C-F" or "Na-Cl")
    """

    # Searching by Database Providers
    aflow = forms.BooleanField(required=False)
    cod = forms.BooleanField(required=False)
    jarvis = forms.BooleanField(required=False)
    materials_project = forms.BooleanField(required=False)
    simmate = forms.BooleanField(required=False)
    oqmd = forms.BooleanField(required=False)

    def clean_chemical_system(self):

        # Our database expects the chemical system to be given in alphabetical
        # order, but we don't want users to recieve errors when they search for
        # "Y-C-F" instead of "C-F-Y". Therefore, we fix that for them here! We
        # also check if the user wants the chemical subsystems included.

        # Grab what the user submitted
        chemical_system = self.cleaned_data["chemical_system"]

        # TODO:
        # Make sure that the chemical system is made of valid elements and
        # separated by hyphens

        # check if the user wants subsystems included (This will be True or False)
        if self.cleaned_data["include_subsystems"]:
            systems_cleaned = get_chemical_subsystems(chemical_system)
        # otherwise just clean the single system
        else:
            # Convert the system to a list of elements
            systems_cleaned = chemical_system.split("-")
            # now recombine the list back into alphabetical order
            systems_cleaned = ["-".join(sorted(systems_cleaned))]
            # NOTE: we call this "systems_cleaned" and put it in a list so
            # that our other methods don't have to deal with multiple cases
            # when running a django filter.
        # now return the cleaned value. Note that this is now a list of
        # chemical systems, where all elements are in alphabetical order.
        return systems_cleaned



# -*- coding: utf-8 -*-

from django_filters import rest_framework as filters

from simmate.database.base_data_types import Thermodynamics as ThermodynamicsTable


class Thermodynamics(filters.FilterSet):
    class Meta:
        model = ThermodynamicsTable
        fields = dict(
            energy=["range"],
            energy_per_atom=["range"],
            energy_above_hull=["range"],
            is_stable=["exact"],
            formation_energy=["range"],
            formation_energy_per_atom=["range"],
        )



# -*- coding: utf-8 -*-

import os

from simmate.toolkit import Structure
from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class TripleProduct(ErrorHandler):
    """
    This error handler swaps the b and c lattice vectors when VASP fails to handle
    its basis vectors properly.
    """

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["ERROR: the triple product of the basis vectors"]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # load the structure and save it as a backup file
        poscar_filename = os.path.join(directory, "POSCAR")
        structure = Structure.from_file(poscar_filename)
        structure.to("POSCAR", poscar_filename + "_original")

        # switch the b and c lattice vector of the structure
        structure.make_supercell([[1, 0, 0], [0, 0, 1], [0, 1, 0]])
        structure.to("POSCAR", poscar_filename)
        correction = "adjusted lattice basis to swap b and c vectors"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

"""
This converter simply imports a pymatgen class. With this you can quickly
convert between Simmate.toolkit and JARVIS.atoms objects.

``` python
from simmate.toolkit import Structure
from simmate.file_converters.structure.jarvis import JarvisAtomsAdaptor


initial_structure = Structure.from_file("example.cif")

# convert to JARVIS
new_jarvis_object = JarvisAtomsAdaptor.get_atoms(structure)

# convert back to PyMatgen/Simmate
new_toolkit_object = JarvisAtomsAdaptor.get_structure(new_jarvis_object)

# alternatively, you can convert back to PyMatgen Molecule
new_molecule_object = JarvisAtomsAdaptor.get_molecule(new_jarvis_object)
```
"""


from pymatgen.io.jarvis import JarvisAtomsAdaptor

__all__ = ["JarvisAtomsAdaptor"]



# -*- coding: utf-8 -*-

"""
This module is for Simmate's test suite. You'll only use this if you are
contributing to the source code and making new tests.

Nearly all of Simmate's tests stem from toolkit objects, so this file loads sample
objects using the `toolkit.base_data_types` module. These Structures and 
Compositions can be used in any test.

Read more on pytest fixtures [here](https://docs.pytest.org/en/6.2.x/fixture.html).
This file helps share fixtures accross files as described 
[here](https://docs.pytest.org/en/6.2.x/fixture.html#conftest-py-sharing-fixtures-across-multiple-files).
"""


import os
import shutil

import pytest
from click.testing import CliRunner
from django.contrib.auth.models import User
from prefect.testing.utilities import prefect_test_harness

from simmate.utilities import get_directory
from simmate.toolkit import Structure, Composition, base_data_types
from simmate.database.base_data_types import Spacegroup
from simmate.website.test_app.models import TestStructure

COMPOSITIONS_STRS = [
    "Fe1",
    "Si2",
    "C4",
    "Ti2O4",
    "Al4O6",
    "Si4N4O2",
    "Si4O8",
    "Sr4Si4N8",
    "Mg4Si4O12",
]


def get_structure_files():
    """
    Lists the full filename paths all of the files in the following directory:
       - toolkit/base_data_types/test/test_structures
    """
    # We want the full path of these filenames, so we that's why there are
    # extra os joins here.
    structure_dir = os.path.join(
        os.path.dirname(base_data_types.__file__),
        "test",
        "test_structures",
    )
    cif_filenames = [os.path.join(structure_dir, f) for f in os.listdir(structure_dir)]
    return cif_filenames


STRUCTURE_FILES = get_structure_files()


@pytest.fixture(scope="session", params=COMPOSITIONS_STRS)
def composition(request):
    """
    Gives a iteratible parameter of example compositions, where the compositions
    are given as ToolkitComposition objects.

    Use this fixture when you want to run a test on all of these compositions
    one at a time. For example, you would run a test like...

    ``` python
    # This function will be ran once for each composition
    def test_example(composition):

        # Do something with your composition.
        # We use a dummy example line here.
        assert composition
    ```
    """
    return Composition(request.param)


@pytest.fixture(scope="session")
def sample_compositions():
    """
    Gives a dictionary of example compositions to use, where the compositions
    are given as ToolkitComposition objects.

    Use this fixture when you want to a specific compositions within a test.
    For example, you would run a test like...

    ``` python
    def test_example(sample_compositions):

        # grab your desired composition
        composition = sample_compositions["Si2"]

        # now run any test you'd like with the object.
        # We use a dummy example line here.
        assert composition == Composition("Si2")
    ```
    """
    return {c: Composition(c) for c in COMPOSITIONS_STRS}


@pytest.fixture(scope="session", params=STRUCTURE_FILES)
def structure(request):
    """
    Gives a iteratible parameter of example structures, where the structures
    are given as ToolkitStructure objects.

    Use this fixture when you want to run a test on all of these structures
    one at a time. For example, you would run a test like...

    ``` python
    # This function will be ran once for each structure
    def test_example(structure):

        # Do something with your structure.
        # We use a dummy example line here.
        assert structure
    ```
    """
    return Structure.from_file(request.param)


@pytest.fixture(scope="session")
def sample_structures():
    """
    Gives a dictionary of example structures to use.

    All of these structures are loaded from files located in...
        simmate/toolkit/base_data_types/test/test_structures

    The structures are given as ToolkitStructure objects and the key are the
    filenames they came from (excluding filename extensions)

    Use this fixture when you want to a specific structures within a test.
    For example, you would run a test like...

    ``` python
    def test_example(sample_structures):

        # grab your desired composition
        structure = sample_structures["C_mp-48_primitive"]

        # now run any test you'd like with the object.
        # We use a dummy example line here.
        assert structure
    ```
    """

    # Now load all of the structures. This is a dictionary that where you
    # can access structures with keys like "SiO2_mp-7029_primitive"
    structures = {
        filename.split(os.path.sep)[-1].strip(".cif"): Structure.from_file(filename)
        for filename in STRUCTURE_FILES
    }

    return structures


@pytest.fixture(scope="session")
def django_db_setup(
    django_db_setup,
    django_db_blocker,
    sample_structures,
    # django_user_model,
):
    """
    This fixture loads test data into the database that can be queried accross
    all other tests. For now, we only add Spacegroups, 10 sample structures,
    and a user.
    """
    with django_db_blocker.unblock():

        # add a user
        User.objects.create_user(
            username="test_user",
            password="test_password",
        )

        # populate spacegroup data
        Spacegroup._load_database_from_toolkit()

        # now add test structures
        for name, structure in sample_structures.items():
            structure_db = TestStructure.from_toolkit(structure=structure)
            structure_db.save()


def copy_test_files(
    tmpdir,
    test_directory: str,
    test_folder: str,
):
    """
    This is a test utility that takes a given directory and copies it's content
    over to a temporary directory. You'll often use this when you want to modify
    files within the test directory (which is often the case with ErrorHandlers).

    Here is an example use-case:
    ``` python
    from somewhere import ExampleHandler
    from simmate.conftest import copy_test_files

    def test_example(tmpdir):

        # Make our temporary directory with copied files
        copy_test_files(
            tmpdir,
            test_directory=__file__,
            test_folder="test_example",
        )

        # then you can do things like...
        error_handler = ExampleHandler()
        error_handler.check(tmpdir)
    ```
    """

    # grab the path to the directory with all the test files
    source_directory = os.path.join(
        os.path.dirname(os.path.abspath(test_directory)),
        test_folder,
    )

    # recursively copy all files in the directory over to the temporary directory
    shutil.copytree(
        src=source_directory,
        dst=tmpdir,
        dirs_exist_ok=True,
    )


def make_dummy_files(*filenames: str):
    """
    This is a utility that creates files. The content of these files are not
    important -- but they are created because sometimes ErrorHanlders may simply
    check to see that a file exists.
    """

    for filename in filenames:
        # make sure the parent dir of the filename exists
        get_directory(os.path.dirname(filename))

        # now make the file
        with open(filename, "w") as file:
            file.write("This is a dummy file for testing.")


@pytest.fixture(scope="session")
def command_line_runner():
    return CliRunner()


@pytest.fixture(autouse=True, scope="session")
def prefect_test_fixture():
    """
    For all prefect flows and tasks, this will automatically use a dummy-database
    """
    with prefect_test_harness():
        yield



# -*- coding: utf-8 -*-

import os
import time

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class Frozen(ErrorHandler):
    """
    Checks when the output file has last been editted. If the job has been sitting
    for a long time (i.e. 1 hour), we consider it frozen and want to change the
    ALGO from Normal to Fast or alternatively reduce SYMPREC.
    """

    is_monitor = True

    def __init__(self, timeout_limit: float = 3600):
        self.timeout_limit = timeout_limit

    def check(self, directory: str) -> bool:
        """
        Check for error in the specified directory. Note, we assume that we are
        checking the vasp.out file. If that file is not present, we say that there
        is no error because another handler will address this.
        """

        # establish the full path to the output file
        output_filename = os.path.join(directory, "vasp.out")

        # check to see that the file is there first
        if os.path.exists(output_filename):

            # check when the file was last editted
            time_last_edit = os.path.getmtime(output_filename)

            # see how long ago this was and if it was longer than our timeout
            if time.time() - time_last_edit > self.timeout_limit:

                # if so, we have a frozen error and need to report it
                return True

        # If the file doesn't exist, we are not seeing any error yet.
        # This line will also be reached if the last edit of the file was under
        # our timeout specified -- where the job is still running and looks good.
        return False

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        #########
        # UNIQUE TO SIMMATE (i.e. changed from original custodian FrozenHandler)
        # if Kerker mixing is present (IMIX=1), then this error was likely caused
        # by the Brmix handler. We therefore delete this setting and retry.
        # See https://github.com/jacksund/simmate/issues/159
        if incar.get("IMIX", None) == 1:
            # delete the CHGCAR and WAVECAR to ensure the next run is a clean start.
            os.remove(os.path.join(directory, "CHGCAR"))
            os.remove(os.path.join(directory, "WAVECAR"))
            incar.pop("IMIX")
            incar.to_file(incar_filename)
            return "Removed IMIX=1 and deleted CHGCAR and WAVECAR"
        #########

        # check what the current ALGO is. If it's not set, that means it's using
        # the default which is "Normal".
        current_algo = incar.get("ALGO", "Normal")
        # also check the SYMPREC, where default is 1e-5
        current_symprec = incar.get("SYMPREC", "1e-5")

        # If the current algo is Fast, then switch it to Normal
        if current_algo == "Fast":
            # Set the new value
            incar["ALGO"] = "Normal"
            # rewrite the new INCAR file
            incar.to_file(incar_filename)
            # return the description of what we did
            return "switched ALGO from Fast to Normal"

        # If the ALGO didn't do the trick, trying setting the SYMPREC to a
        # specific value.
        elif current_symprec != 1e-8:
            # Set the new value
            incar["SYMPREC"] = 1e-8
            # rewrite the new INCAR file
            incar.to_file(incar_filename)
            # return the description of what we did
            return f"switched SYMPREC from {current_symprec} to 1e-8"

        # If none of the above worked, then we were not able to fix the error
        else:
            raise Exception("Unable to fix Frozen error")



# -*- coding: utf-8 -*-

import itertools
from typing import List

from pymatgen.core import Species
from pymatgen.analysis.structure_prediction.substitutor import Substitutor

from simmate.toolkit import Composition, Structure
from .known import get_known_structures


def get_structures_from_substitution_of_known(
    composition: Composition,
    **kwargs,  # passed to get_known_structures
) -> List[Structure]:
    """
    Predicts the most likely element substitutions for a composition, and then
    based off the predictions, it will go through all database tables in the
    `simmate.database.third_parties` module generate substituted structures.

    Each database table must be populated -- otherwise this function will
    return an empty list.
    """

    # make sure the composition has oxidation states and if not, add states
    # using the built-in guessing method
    if not isinstance(composition.elements[0], Species):
        oxi_composition = composition.add_charges_from_oxi_state_guesses()
    else:
        # make a copy just in case we perform manipulations below
        oxi_composition = composition.copy()

    # and grab the list of species, which we will need later
    elements_final = [str(element) for element in oxi_composition]
    oxi_states = {
        element.symbol: element.oxi_state for element in oxi_composition.elements
    }

    # init the substitution engine and grab the list of likely substituitions
    substitutor = Substitutor()

    # ----------
    # BUG: This code should be a single line:
    #   expected_substituitions = substitutor.pred_from_comp(oxi_composition)
    # Howevever, pymatgen's function fails when even a single element is not
    # allowed with their dataset (e.g. any carbides will fail). To get around
    # this, I use pred_from_list instead of pred_from_comp, and I keep trying
    # until I get one that works. The while loop here is try every combination
    # of elements (starting with all of them and then try smaller combos) and
    # exits as soon as I find one that works.
    nelements = len(oxi_composition.elements)
    has_subs = False
    while (not has_subs) and (nelements > 0):
        for elements_to_sub in itertools.combinations(
            oxi_composition.elements, nelements
        ):
            try:
                expected_substituitions = substitutor.pred_from_list(elements_to_sub)
                break  # exit immediately once successful
            except:
                expected_substituitions = None
        if expected_substituitions:
            has_subs = True
        else:
            # try the cycle again but with one less element in the combinations
            nelements -= 1

    if not expected_substituitions:
        raise Exception("Failed to find viable substituitions for this composition")

    # Confirm charge balance for most-likely structures
    # This loop is copy/pasted from...
    # https://github.com/materialsproject/pymatgen/blob/e0c1ee845e7131213055c5b4fc6df2cb12fcc667/pymatgen/analysis/structure_prediction/substitutor.py#L238
    output = []
    for p in expected_substituitions:
        subs = p["substitutions"]
        charge = 0
        for i_el in oxi_composition.elements:
            f_el = subs[i_el]
            charge += f_el.oxi_state * oxi_composition[i_el]
        if charge == 0:
            output.append(p)
    expected_substituitions = output

    # ----------

    # Keep a list of all structures that we create
    structures_final = []

    # now iterate through the expected substitutions, grab the known structures
    # from databases, and then perform the substitution to generate new structures
    for substituition in expected_substituitions:

        # The methods below expect this sub dictionary to be based off of strings
        # instead of Species objects, so we convert them first.
        substituition_cleaned = {
            str(element_new): str(element_orig)
            for element_new, element_orig in substituition["substitutions"].items()
        }
        # this is the same as above with keys/values flipped
        substituition_cleaned_2 = {
            str(element_orig): str(element_new)
            for element_new, element_orig in substituition["substitutions"].items()
        }

        # Generate the composition that we will use for source structures
        composition_source = oxi_composition.replace(substituition_cleaned.copy())
        # BUG: this method is destructive to the input dictionary so we must
        # pass a copy instead
        print([oxi_composition, composition_source, substituition_cleaned])

        # Now grab all known structures with this composition
        source_structures = get_known_structures(composition_source, **kwargs)

        # skip to the next substitution type if no structures were found
        if not source_structures:
            continue

        # grab the oxidation states from our original substituition
        oxi_states_sub = {
            element_orig.symbol: element_orig.oxi_state
            for element_new, element_orig in substituition["substitutions"].items()
        }
        # in case oxi_states_sub does not contain all of the elements of the
        # composition (see the itertools.combinations loop above), then we
        # need to add back the oxi state of missing elements here.
        for key, value in oxi_states.items():
            if key not in oxi_states_sub:
                oxi_states_sub[key] = value

        # decorate the structure with these oxi states
        for structure in source_structures:
            structure.add_oxidation_state_by_element(oxi_states_sub)

        # BUG: this is the method that's built-in with the Substitutor class,
        # but it throws random errors and also is more code... I do away with
        # this for now, but keep the code fore reference.
        #
        # finally, use this list of structures to generate sub structures. this
        # method requires the input structures to be given as a dictionary
        # structure_list = [
        #     {"structure": structure, "id": structure.database_object.id}
        #     for structure in source_structures
        # ]
        # new_structures = substitutor.pred_from_structures(
        #     target_species=elements_final,
        #     structures_list=structure_list,
        # )
        # # clean these new structures
        # for structure in new_structures:
        #     # note, structure is a TransformedStructure obj to start
        #     structure_final = structure.final_structure
        #     structure_final.remove_oxidation_states()
        #     structures_final.append(structure_final)

        for structure in source_structures:
            new_structure = structure.copy()
            new_structure.replace_species(substituition_cleaned_2)
            # TODO: scale structures to reasonable volume..?
            structures_final.append(new_structure)

    return structures_final



# -*- coding: utf-8 -*-

from simmate.workflow_engine import task, Workflow


@task
def run_customized_s3task(
    workflow_base: Workflow,
    input_parameters: dict,
    updated_settings: dict,
):

    # ensure this workflow has a S3Task class attribute
    if not hasattr(workflow_base, "s3task"):
        raise Exception(
            "Dynamically updating settings is only supported for workflows "
            "generated via the `s3task_to_workflow` utility."
        )

    # we can't just copy the workflow and update attributes, as this will update
    # attributes of source workflow as well. Instead, we must subclass the underlying
    # tasks and then update their attributes. Therefore, this process below seeks
    # to copy and update relevant attributes before subclassing at the end.
    #   see https://stackoverflow.com/questions/9541025/

    # we start with the original s3task class and keep track of attributes we
    # need to update.
    s3task = workflow_base.s3task
    final_attributes = {}

    # Key will be something like "incar" and the update_values will be a
    # dictionary such as {"ENCUT": 520, "NPAR": 3}
    for update_attribute, update_values in updated_settings.items():

        original_config = getattr(s3task, update_attribute)

        # we will be deleting/editting the dictionary so we need to make a copy
        new_config = original_config.copy()

        # If a user sets EDIFF = 1e-3 and there is EDIFF__per_atom in the
        # incar dictionary, we need to remove that key as well. We therefore
        # remove all other "examplekey__*" settings from the dictionary.
        for update_key in update_values.keys():
            for subkey in original_config.keys():
                if subkey.startswith(f"{update_key}__"):
                    new_config.pop(subkey)
        # BUG: what about keywords like multiple_keywords__smart_ldau? I may need
        # to refactor the INCAR class to handle this.

        # now we can add the custom values
        new_config.update(update_values)

        final_attributes[update_attribute] = new_config

    # Dynamically create a subclass of the original s3task
    NewS3Task = type(f"Custom{s3task.__name__}", (s3task,), final_attributes)

    # now run the task with the remaining parameters
    # Note, we are using run_config instead of run because we do not want this
    # registered as a task run of the original class
    result = NewS3Task.run_config(**input_parameters)

    return result



# -*- coding: utf-8 -*-

import itertools

from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from simmate.toolkit import Composition, Structure
from simmate.database import connect
from simmate.database.third_parties import AflowPrototype


class FromAflowPrototypes:
    """
    Generates structures of a target composition by using AFLOW prototypes as
    templates.

    # EXACT MODE
    Given a compositon, this class can be used to create structures from
    the **exactly-matching** prototypes in the AFLOW library. This means we
    are only using prototypes that have the same anonymous formula. For example,
    if you request structures for Ca2N to be made, only prototypes that have
    the AB2 anonymoous formula will be used (such as the MoS2 prototype).

    # ORDERED MODE
    Allow prototypes such as PbClF to be used on compositions like Ca2N
    (i.e. 3 wyckoff sites available and 3 sites to to create). Prototypes are
    selected based on wyckoff site multiplicity.

    # DISORDERED MODE
    (Not implemented yet)
    At the moment, this does NOT allow for disordered versions of prototypes.
    This would be like using a supercell of a 1-site prototype for a 2-site target
    composition. A concrete example of this is using the NaCl rocksalt prototype
    for a NaClBr or NaKClBr compositions -- each of these compositions would
    require a supercell of the NaCl prototype structure becuase they need more
    than 2 sites.
    """

    def __init__(
        self,
        composition: Composition,
        allow_multiples: bool = True,
        max_sites: int = 115,
        mode: str = "exact",  # options are "exact" and "ordered"
    ):
        """
        #### Parameters

        - `composition`:
            The desired composition to create structures for

        - `allow_multiples`:
            Whether to allow structures where the reduced compositions are the
            same, but the actual number of sites differs. For example, Ca6N3
            would be allowed to return structures that give Ca2N, Ca8N4, and
            so forth.

        - `max_sites`:
            Maximum sites allowed in the created structures. This defaults to
            115 because this would include all prototypes in the aflow database
            (currently, the largest is prototype is 105 sites)

        - `mode`:
            The method to use for selecting prototypes. Options are "exact" and
            "ordered" (see class docstring above for what each does). Default
            is "exact"
        """

        # Store these input properties for reference
        self.composition = composition
        self.max_sites = max_sites
        self.allow_multiples = allow_multiples
        self.mode = mode

        # Load the template prototypes from the database. These methods set
        # the self.prototype_structures attribute that will be used below
        if mode not in ["exact", "ordered"]:
            raise Exception(f"mode provided ({mode}) is not allowed")
        elif mode == "exact":
            self._get_exact_prototypes()
            self._create_structures_from_exact()
        elif mode == "ordered":
            if self.allow_multiples:
                raise Exception(
                    "using allow_multiples=True with mode='ordered' often results "
                    "in millions of structures, so this input is currently not allowed."
                )
            self._get_ordered_prototypes()
            self._create_structures_from_ordered()
        # TODO: Consider breaking this into subclasses that inherit from a base
        # class. Alternatively, keep these here and allow combination of modes.

        # ---------------------------------------------------------------------

        # We now have all of the structures, but they will have unreasonable
        # volume and atomic distances right now. We iterate through each
        # and try to fix this by doing the following:
        #   (1) Scaling the structure to a predicted volume
        #   (2) Sanitizing the structure
        #   (3) [TODO] scaling the lattice vectors to avoid atoms being too close
        volume_guess = self.composition.volume_estimate()
        for i, structure in enumerate(self.structures):
            structure.scale_lattice(volume_guess)
            structure_santized = structure.get_sanitized_structure()
            self.structures[i] = structure_santized

        # ---------------------------------------------------------------------

        # lastly, go through and remove duplicate structures
        unique_structures = []
        for structure in self.structures:
            if structure not in unique_structures:
                unique_structures.append(structure)
        self.structures = unique_structures
        # OPTIMIZE: should I do a more robust check for duplicate structures?
        # For example, I could use a fingerprint validator.

    def _get_exact_prototypes(self):
        # Load all of the prototype structures that have a compatible number of
        # sites in the structure. For this we require that the prototypes
        # meet the following:
        #   (1) prototypes have the same anonymous formula as target the composition
        #   (2) the number of wykcoff sites is at least the number of target elements
        # Note: #1 ensures that #2 is true here, so we don't actually need to
        # check for both -- just #1.

        # We allow for multiples so we want all multiples nsites in our
        # desired composition. For example if our compositon, has 6 sites, we
        # would pull all prototypes that have 6, 12, 19, ..., 6*N sites.
        # First generate a list of the allow number of sites
        if self.allow_multiples:
            target_nsites_list = [
                int(self.composition.num_atoms * n)
                for n in range(
                    1, (self.max_sites // int(self.composition.num_atoms)) + 1
                )
            ]
            # Now filter off the prototypes that have the proper number of sites
            prototype_entries = AflowPrototype.objects.filter(
                formula_anonymous=self.composition.anonymized_formula,
                nsites__in=target_nsites_list,
            ).all()
        else:
            prototype_entries = AflowPrototype.objects.filter(
                formula_anonymous=self.composition.anonymized_formula,
                nsites=int(self.composition.num_atoms),
            ).all()

        # for user convience, let's store these selected prototypes as a dataframe
        # and attach it to this class instance
        self.prototype_dataframe = prototype_entries.to_dataframe()

        # Now that we have the prototype entries, let's convert them to
        # structure objects for us to use elsewhere.
        self.prototype_structures = prototype_entries.to_toolkit()

    def _get_ordered_prototypes(self):

        # Load all of the prototype structures that have a compatible
        # number of sites in the structure.

        # There are two conditions that we need to meet for prototypes here:
        #   (1) the number of wyckoff sites is compatible with nelements OR nsites
        #   (2) the number of individual sites is compatible with nsites
        # This is because we can only insert at symmetrically unique sites (bc
        # we ignore disoreder superlattices here) and the total number of sites
        # after symmetry operations needs to match our target composition.
        # Note that nsites is *not* always a multiple of nsites_wyckoff! For
        # example the Spinel prototype has nsites=14 and nsites_wyckoff=3.

        # Generate a list of the allowed number of wyckoff sites
        # We use 40 here because this the maximum number of unique wyckoff
        # sites of any prototype (actual max is 36).
        target_nsites_wyckoff_list = set(
            [
                int(self.composition.num_atoms * n)
                for n in range(1, (40 // len(self.composition.elements)) + 1)
            ]
            + [
                int(len(self.composition.elements) * n)
                for n in range(1, (40 // int(self.composition.num_atoms)) + 1)
            ]
        )

        # If we allow for multiples then  want all multiples nsites in our
        # desired composition. For example if our compositon, has 6 sites, we
        # would pull all prototypes that have 6, 12, 18, ..., 6*N sites.
        if self.allow_multiples:
            # First generate a list of the allow number of sites
            target_nsites_list = [
                int(self.composition.num_atoms * n)
                for n in range(
                    1, (self.max_sites // int(self.composition.num_atoms)) + 1
                )
            ]
            # Now filter off the prototypes that have the proper number of sites
            prototype_entries = AflowPrototype.objects.filter(
                nsites_wyckoff__in=target_nsites_wyckoff_list,
                nsites__in=target_nsites_list,
            ).all()

        # If we don't want multiples, we can just filter by
        # an exact match for number of total sites and wyckoff sites
        else:
            prototype_entries = AflowPrototype.objects.filter(
                nsites_wyckoff__in=target_nsites_wyckoff_list,
                nsites=int(self.composition.num_atoms),
            ).all()

        # for user convience, let's store these selected prototypes as a dataframe
        # and attach it to this class instance
        self.prototype_dataframe = prototype_entries.to_dataframe()

        # Now that we have the prototype entries, let's convert them to
        # structure objects for us to use below.
        # Note, becuase the ordered method relies on wyckoff sites, we want
        # these as symmetrized structures.
        self.prototype_structures = [
            SpacegroupAnalyzer(prototype).get_symmetrized_structure()
            for prototype in prototype_entries.to_toolkit()
        ]

    def _create_structures_from_exact(self):
        # We now generate all prototypes here. We store them in a final list
        # as we make them too.
        self.structures = []
        for prototype in self.prototype_structures:

            # we need to match elements in our target composition to those in
            # the prototype. For example, Ca2NF with the NaCrS2 prototype would
            # need to match S-->Ca, Cr-->N, and Na-->F.
            #
            # To do this, we need to look at every possible order of target
            # elements. We then iterate through these combinations and check
            # which ones are valid. For example, a Ca2NCl would give us possible
            # orderings of Ca-N-F, Ca-F-N, N-Ca-F, N-F-Ca, F-Ca-N, and F-N-Ca.
            # We then go through each ordering and see which ones will give our
            # target composition when substituted for Na-Cr-S in order.
            for target_elements in itertools.permutations(self.composition.elements):

                # Using this ordering of target elements we can now
                # attempt the subsitution for the prototype

                # We first sure to make a copy of the prototype that we can alter.
                new_structure = prototype.copy()

                # Go through and substitute out the elements
                for target_element, current_element in zip(
                    target_elements, prototype.composition
                ):
                    new_structure[current_element] = target_element

                # Now let's check to see that the final composition matches
                # what we are trying to produce. If so, we store the structure!
                # Note, we are after the reduce composition because we want
                # to allow things like Ca2NF == Ca4N2F2 as True.
                if (
                    new_structure.composition.reduced_composition
                    == self.composition.reduced_composition
                ):
                    self.structures.append(new_structure)

    def _create_structures_from_ordered(self):

        # Create a flattened list of elements (e.g. Ca2N --> [Ca, Ca, N])
        # Note we use the reduced composition to ensure the smallest list, and
        # we duplicate this list to the proper size based on the prototype
        # being used.
        element_list = self.composition.reduced_composition.elements
        element_list_expanded = []
        for element in self.composition.reduced_composition.elements:
            element_list_expanded += [element] * int(self.composition[element])

        self.structures = []
        for prototype in self.prototype_structures:

            if len(prototype.equivalent_indices) % len(element_list_expanded) == 0:
                mult = len(prototype.equivalent_indices) // len(element_list_expanded)
                element_list_scaled = element_list_expanded * mult
            elif len(prototype.equivalent_indices) % len(element_list) == 0:
                mult = len(prototype.equivalent_indices) // len(element_list)
                element_list_scaled = element_list * mult
            else:
                raise Exception("Unknown scaling for prototype wyckoffs")

            for target_elements in itertools.permutations(element_list_scaled):

                new_structure = prototype.copy()

                # Go through and substitute out the elements
                for target_element, current_indicies in zip(
                    target_elements, prototype.equivalent_indices
                ):
                    for current_index in current_indicies:
                        new_structure.replace(current_index, target_element)

                # Now let's check to see that the final composition matches
                # what we are trying to produce. If so, we store the structure!
                # Note, we are after the reduce composition because we want
                # to allow things like Ca2NF == Ca4N2F2 as True.
                if (
                    new_structure.composition.reduced_composition
                    == self.composition.reduced_composition
                ):
                    # convert from symmetrized structure back to toolkit structure
                    new_structure_toolkit = Structure(
                        lattice=new_structure.lattice,
                        species=new_structure.species,
                        coords=new_structure.frac_coords,
                    )
                    self.structures.append(new_structure_toolkit)



# -*- coding: utf-8 -*-

import os

from simmate.toolkit import Structure
from simmate.calculators.vasp.inputs import Incar, Poscar, Kpoints, Potcar
from simmate.calculators.vasp.tasks.relaxation import MITRelaxation


class MITDynamics(MITRelaxation):
    """
    This task is a reimplementation of pymatgen's
    [MITMDSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MITMDSet).

    Runs a molecular dynamics simulation using MIT Project settings. The lattice
    will remain fixed during the run.

    By default, this will run from 300-1200 Kelvin over 10,000 steps of 2
    femtoseconds, but start/end temperatures as well as step size/number can
    be adjusted when initializing this class. Note, setting parameters in the
    init is atypical for Simmate tasks, but we allow it for MD run because it
    does not affect the interopability of results -- that is, results can
    be compared accross runs regardless of temp/time.

    Provide your structure as the desired supercell, as the setup of this
    calculation does not modify your input structure.
    """

    confirm_convergence = False

    incar = MITRelaxation.incar.copy()
    incar.update(
        dict(
            # Unique to this task, we want to allow users to set these temperatures
            # but to keep with Simmate's strategy of showing all settings up-front,
            # we set these messages here.
            # TODO: consider making a "__user_input" incar tag that accepts a default
            TEBEG="Defaults to 300 but can be set by the user",  # start temperature
            TEEND="Defaults to 1200 but can be set by the user",  # end temperature
            POTIM="Defaults to 2 but can be set by the user",  # time step (in fs)
            NSW="Defaults to 10,000 but can be set by the user",  # number of steps
            #
            EDIFF__per_atom=1e-06,
            LSCALU=False,
            LCHARG=False,
            LPLANE=False,
            LWAVE=True,
            ISMEAR=0,
            NELMIN=4,
            LREAL=True,
            BMIX=1,
            MAXMIX=20,
            NELM=500,
            NSIM=4,  # same as VASP default but pymatgen sets this
            ISYM=0,  # turn off symmetry
            ISIF=0,  # only update atom sites; lattice is fixed; no lattice stress
            IBRION=0,  # turns on molecular dynamics
            KBLOCK=100,
            SMASS=0,
            ISPIN=1,  # pymatgen makes this a kwarg but we fix it to pmg's default
            PREC="Low",
            NBLOCK=1,  # same as VASP default but pymatgen sets this
        )
    )
    incar.pop("MAGMOM__smart_magmom")
    incar.pop("multiple_keywords__smart_ldau")
    incar.pop("EDIFF")
    incar.pop("ENCUT")

    # For now, I turn off all error handlers.
    # TODO
    error_handlers = []

    @classmethod
    def setup(
        cls,
        structure: Structure,
        directory: str,
        temperature_start: int = 300,
        temperature_end: int = 1200,
        time_step: float = 2,
        nsteps: int = 10000,
        **kwargs,
    ):

        # run cleaning and standardizing on structure (based on class attributes)
        structure_cleaned = cls._get_clean_structure(structure, **kwargs)

        # write the poscar file
        Poscar.to_file(structure_cleaned, os.path.join(directory, "POSCAR"))

        # Combine our base incar settings with those of our parallelization settings
        # and then write the incar file. Note, we update the values of this incar,
        # so we make a copy of the dict.
        incar = cls.incar.copy()
        incar["TEBEG"] = temperature_start
        incar["TEEND"] = temperature_end
        incar["POTIM"] = time_step
        incar["NSW"] = nsteps
        incar = Incar(**incar) + Incar(**cls.incar_parallel_settings)
        incar.to_file(
            filename=os.path.join(directory, "INCAR"),
            structure=structure,
        )

        # if KSPACING is not provided in the incar AND kpoints is attached to this
        # class instance, then we write the KPOINTS file
        if cls.kpoints and ("KSPACING" not in cls.incar):
            Kpoints.to_file(
                structure,
                cls.kpoints,
                os.path.join(directory, "KPOINTS"),
            )

        # write the POTCAR file
        Potcar.to_file_from_type(
            structure.composition.elements,
            cls.functional,
            os.path.join(directory, "POTCAR"),
            cls.potcar_mappings,
        )



# -*- coding: utf-8 -*-

from simmate.toolkit import Structure
from simmate.workflow_engine import Workflow
from simmate.workflow_engine.common_tasks import (
    load_input_and_register,
    save_result,
)
from simmate.calculators.vasp.tasks.dynamics import MITDynamics
from simmate.calculators.vasp.database.dynamics import MITDynamicsRun


class Dynamics__Vasp__Mit(Workflow):
    description_doc_short = "uses MIT Project settings"
    database_table = MITDynamicsRun
    s3task = MITDynamics

    @classmethod
    def run_config(
        cls,
        structure: Structure,
        command: str = None,
        source: dict = None,
        directory: str = None,
        #
        # extra parameters unique to molecular dynamics runs
        copy_previous_directory: bool = False,
        temperature_start: float = 300,
        temperature_end: float = 1200,
        time_step: float = 2,
        nsteps: int = 10000,
    ):
        parameters_cleaned = load_input_and_register(
            structure=structure,
            source=source,
            directory=directory,
            command=command,
            copy_previous_directory=copy_previous_directory,
            temperature_start=temperature_start,
            temperature_end=temperature_end,
            time_step=time_step,
            nsteps=nsteps,
        )

        result = cls.s3task.run(**parameters_cleaned).result()
        calculation_id = save_result(result)

        return result



# -*- coding: utf-8 -*-

from simmate.utilities import get_doc_from_readme

__doc__ = get_doc_from_readme(__file__)

from simmate.calculators.vasp.database.relaxation import (
    MatprojRelaxation,
    MatprojIonicStep,
    MITRelaxation,
    MITIonicStep,
    Quality00Relaxation,
    Quality00IonicStep,
    Quality01Relaxation,
    Quality01IonicStep,
    Quality02Relaxation,
    Quality02IonicStep,
    Quality03Relaxation,
    Quality03IonicStep,
    Quality04Relaxation,
    Quality04IonicStep,
    StagedRelaxation,
    MatVirtualLabCINEBEndpointRelaxation,
    MatVirtualLabCINEBEndpointIonicStep,
)

from simmate.calculators.vasp.database.energy import (
    MatprojStaticEnergy,
    MITStaticEnergy,
    Quality04StaticEnergy,
    NEBEndpointStaticEnergy,
)

from simmate.calculators.vasp.database.band_structure import (
    MatprojBandStructure,
)

from simmate.calculators.vasp.database.density_of_states import (
    MatprojDensityOfStates,
)

from simmate.calculators.vasp.database.nudged_elastic_band import (
    MITDiffusionAnalysis,
    MITMigrationHop,
    MITMigrationImage,
)

from simmate.calculators.vasp.database.dynamics import (
    MITDynamicsRun,
    MITDynamicsIonicStep,
)

from simmate.calculators.vasp.database.population_analysis import (
    MatprojBaderAnalysis,
    MatprojBaderELFAnalysis,
)

from simmate.calculators.vasp.database.customized import (
    CustomizedVASPCalculation,
)

from simmate.toolkit.structure_prediction.evolution.database import (
    EvolutionarySearch,
    StructureSource,
)



# -*- coding: utf-8 -*-

from io import BytesIO
import base64

from django import template
from django.utils.safestring import mark_safe

# We need a registration instance in order to configure everything with Django
register = template.Library()


@register.filter(name="matplotlib_figure", is_safe=True)
def matplotlib_to_html(plot):
    """
    Converts a matplotlib figure into an html element.
    """

    # Lmao I have no idea how this works and I just got lucky when following
    # these stack overflow posts... I should switch to plotly ASAP so I can
    # remove this filter.
    #   https://stackoverflow.com/questions/14824522/
    #   https://stackoverflow.com/questions/40534715/
    # Alternatively, I could rewrite this function to save a randomly generated
    # file to the /static/runtime folder and then passing back an html <img> tag
    # that points back to this image.

    figdata = BytesIO()
    plot.savefig(figdata, format="png")
    figdata.seek(0)
    data = figdata.getvalue()
    data_decoded = base64.b64encode(data).decode("utf-8").replace("\n", "")
    html = f"""
    <img id="ItemPreview" class="img-fluid" src="data:image/png;base64,{data_decoded}">
    """

    # BUG: every time this filter is called, I recieve the following warning:
    #
    # .../lib/python3.10/site-packages/pymatgen/util/plotting.py:48: UserWarning:
    # Starting a Matplotlib GUI outside of the main thread will likely fail.
    # plt.figure(figsize=(width, height), facecolor="w", dpi=dpi)
    #
    # This doesn't seem to cause any issues, so I ignore it for now.

    # Because we added new html to our script, we need to have Django check it
    # ensure it safe before returning. Read more about this here:
    # https://docs.djangoproject.com/en/3.2/howto/custom-template-tags/#filters-and-auto-escaping
    return mark_safe(html)



# -*- coding: utf-8 -*-

import os

from pymatgen.electronic_structure.plotter import DosPlotter

from simmate.calculators.vasp.tasks.static_energy.materials_project import (
    MatprojStaticEnergy,
)


class VaspDensityOfStates(MatprojStaticEnergy):
    """
    A base class for density of states (DOS) calculations. This is not meant
    to be used directly but instead should be inherited from.

    This is also a non self-consistent field (non-SCF) calculation and thus uses
    the a fixed charge density from a previous static energy calculation.
    """

    required_files = MatprojStaticEnergy.required_files + ["CHGCAR"]

    @staticmethod
    def _write_output_summary(directory, vasprun):
        """
        In addition to writing the normal VASP output summary, this also plots
        the DOS to "density_of_states.png"
        """

        # run the normal output
        MatprojStaticEnergy._write_output_summary(directory, vasprun)

        # and then generate a DOS plot
        plotter = DosPlotter()

        # Add the total density of States
        plotter.add_dos("Total DOS", vasprun.complete_dos)

        # add element-projected density of states
        plotter.add_dos_dict(vasprun.complete_dos.get_element_dos())

        # If I want plots for individual orbitals
        # for site in vasprun.final_structure:
        #     spd_dos = vasprun.complete_dos.get_site_spd_dos(site)
        #     plotter.add_dos_dict(spd_dos)

        # NOTE: get_dos_dict may be useful in the future

        plot = plotter.get_plot()
        plot_filename = os.path.join(directory, "density_of_states.png")
        plot.savefig(plot_filename)



# -*- coding: utf-8 -*-

from .diffusion.all import (
    Diffusion__Vasp__NebAllPaths,
    Diffusion__Vasp__NebSinglePath,
    Diffusion__Vasp__NebFromEndpoints,
    Diffusion__Vasp__NebFromImages,
)

from .electronic_structure.all import (
    ElectronicStructure__Vasp__MatprojFull,
)

from .population_analysis.all import (
    PopulationAnalysis__Vasp__BaderMatproj,
    PopulationAnalysis__Vasp__ElfMatproj,
    PopulationAnalysis__Vasp__BadelfMatproj,
)

from .static_energy.all import (
    StaticEnergy__Vasp__Matproj,
    StaticEnergy__Vasp__Mit,
    StaticEnergy__Vasp__Quality04,
    StaticEnergy__Vasp__NebEndpoint,
)

from .relaxation.all import (
    Relaxation__Vasp__Matproj,
    Relaxation__Vasp__Mit,
    Relaxation__Vasp__NebEndpoint,
    Relaxation__Vasp__Quality00,
    Relaxation__Vasp__Quality01,
    Relaxation__Vasp__Quality02,
    Relaxation__Vasp__Quality03,
    Relaxation__Vasp__Quality04,
    Relaxation__Vasp__Staged,
)

from .customized import Customized__Vasp__UserConfig

from .dynamics import Dynamics__Vasp__Mit



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class Edddav(ErrorHandler):
    """
    When the gradient is not orthogonal, we make a fix depending on what IMSEAR
    is set to.
    """

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["EDWAV: internal error, the gradient is not orthogonal"]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        correction = ""
        # Delete the CHGCAR if this is a "from-scratch" calc
        current_icharg = incar.get("ICHARG", 0)
        if current_icharg < 10:
            os.remove(os.path.join(dir, "CHGCAR"))
            correction += "deleted CHGCAR and "
        incar["ALGO"] = "All"
        correction += "switched ALGO to Normal"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-


class DriftErrorHandler(ErrorHandler):
    """
    Corrects for total drift exceeding the force convergence criteria.
    """

    def __init__(self, max_drift=None, to_average=3, enaug_multiply=2):
        """
        Initializes the handler with max drift
        Args:
            max_drift (float): This defines the max drift. Leaving this at the default of None gets the max_drift from
                EDFIFFG
        """

        self.max_drift = max_drift
        self.to_average = int(to_average)
        self.enaug_multiply = enaug_multiply

    def check(self):
        """
        Check for error.
        """
        incar = Incar.from_file("INCAR")
        if incar.get("EDIFFG", 0.1) >= 0 or incar.get("NSW", 0) == 0:
            # Only activate when force relaxing and ionic steps
            # NSW check prevents accidental effects when running DFPT
            return False

        if not self.max_drift:
            self.max_drift = incar["EDIFFG"] * -1

        try:
            outcar = Outcar("OUTCAR")
        except Exception:
            # Can't perform check if Outcar not valid
            return False

        if len(outcar.data.get("drift", [])) < self.to_average:
            # Ensure enough steps to get average drift
            return False

        curr_drift = outcar.data.get("drift", [])[::-1][: self.to_average]
        curr_drift = np.average([np.linalg.norm(d) for d in curr_drift])
        return curr_drift > self.max_drift

    def correct(self):
        """
        Perform corrections.
        """
        backup(VASP_BACKUP_FILES)
        actions = []
        vi = VaspInput.from_directory(".")

        incar = vi["INCAR"]
        outcar = Outcar("OUTCAR")

        # Move CONTCAR to POSCAR
        actions.append(
            {"file": "CONTCAR", "action": {"_file_copy": {"dest": "POSCAR"}}}
        )

        # First try adding ADDGRID
        if not incar.get("ADDGRID", False):
            actions.append({"dict": "INCAR", "action": {"_set": {"ADDGRID": True}}})
        # Otherwise set PREC to High so ENAUG can be used to control Augmentation Grid Size
        elif incar.get("PREC", "Accurate").lower() != "high":
            actions.append({"dict": "INCAR", "action": {"_set": {"PREC": "High"}}})
            actions.append(
                {
                    "dict": "INCAR",
                    "action": {"_set": {"ENAUG": incar.get("ENCUT", 520) * 2}},
                }
            )
        # PREC is already high and ENAUG set so just increase it
        else:
            actions.append(
                {
                    "dict": "INCAR",
                    "action": {
                        "_set": {
                            "ENAUG": int(incar.get("ENAUG", 1040) * self.enaug_multiply)
                        }
                    },
                }
            )

        curr_drift = outcar.data.get("drift", [])[::-1][: self.to_average]
        curr_drift = np.average([np.linalg.norm(d) for d in curr_drift])
        VaspModder(vi=vi).apply_actions(actions)
        return {
            "errors": "Excessive drift {} > {}".format(curr_drift, self.max_drift),
            "actions": actions,
        }



# -*- coding: utf-8 -*-

from .load_input_and_register import load_input_and_register
from .save_result import save_result
from .parse_multi_command import parse_multi_command
from .customized_s3task import run_customized_s3task



# -*- coding: utf-8 -*-

from pymatgen.analysis.elasticity.strain import Strain, convert_strain_to_deformation

from simmate.toolkit.transformations.base import Transformation
from simmate.toolkit.creators.vector import UniformlyDistributedVectors


class LatticeStrain(Transformation):

    #!!! should I mutate atomic sites too? USPEX is unclear if they do this in addition to lattice strain
    #!!! if I decide to do this, look at CoordinateMutation class below

    # strain is applied to the lattice and then the lattice is scaled back to target volume
    # https://uspex-team.org/static/file/CPC-USPEX-2006.pdf

    # It looks like structure.apply_strain() is simply scaling each lattice vector and not exactly what we'd like
    # Instead, I found pymatgen.analysis.elasticity.strain module
    # https://pymatgen.org/pymatgen.analysis.elasticity.strain.html
    # Here, my understanding is that a strain matrix is not equivalent to a deformation matrix
    # therefore, I will create a Strain object, then convert it Deformation object, then apply it to the structure

    io_scale = "one_to_one"
    ninput = 1
    use_multiprocessing = False

    def __init__(self, fixed_volume):

        # after straining the lattice, we need to scale it to a fixed volume
        self.fixed_volume = fixed_volume

        # following along with USPEX paper, we need to first establish boundries for the symmetric strain matrix
        # when generating components for this matrix, they randomly select values between -1 and 1 in a guassian distribution
        self.component_generator = UniformlyDistributedVectors(
            min_value=-1, max_value=1
        )

    def apply_transformation(self, structure, max_attempts=100):

        # first we need 6 strain matrix components
        components = self.component_generator.new_vector(size=6)

        # next we need to assemble the strain matrix using these components
        #!!! is there a better way to do this? I'm just using equation (3) in the USPEX paper
        strain_matrix = [
            [1 + components[0], components[5] / 2, components[4] / 2],
            [components[5] / 2, 1 + components[1], components[3] / 2],
            [components[4] / 2, components[3] / 2, 1 + components[2]],
        ]

        # now let's switch to pymatgen functionality

        # first make the Strain object
        strain = Strain(strain_matrix)

        # now convert it to a deformation object
        #!!! do I want the shape to be 'upper' or 'symmetric'??
        deformation = convert_strain_to_deformation(strain, shape="symmetric")

        # deform the structure
        new_structure = deformation.apply_to_structure(structure)

        # scale the lattice volume back to a fixed value
        new_structure.scale_lattice(self.fixed_volume)

        return new_structure



# -*- coding: utf-8 -*-

"""

> :warning: This file is only for use by the Simmate team. Users should instead
access data via the load_remote_archive method.

This file is for pulling JARVIS data into the Simmate database. 

JARVIS has a python package "jarvis-tools" that let's us pull some of their
database dumps. For instructions on how to do this, they provided 
[this link](https://colab.research.google.com/github/knc6/jarvis-tools-notebooks/blob/master/jarvis-tools-notebooks/Get_JARVIS_DFT_final_structures_in_ASE_or_Pymatgen_format.ipynb)

Alternatively, we could manually download 
[their database json files](https://jarvis-materials-design.github.io/dbdocs/thedownloads/). We specifically look at the "3D-materials curated data".

"""

from django.db import transaction

from tqdm import tqdm
from simmate.toolkit import Structure

from simmate.database.third_parties import JarvisStructure


@transaction.atomic
def load_all_structures():
    """
    Only use this function if you are part of the Simmate dev team!

    Loads all structures directly for the JARVIS database into the local
    Simmate database.
    """

    # Jarvis is not a dependency of simmate, so make sure you install it before using
    # this module
    try:
        from jarvis.db.figshare import data as jarvis_helper
    except:
        raise ModuleNotFoundError(
            "You must install jarvis with `conda install -c conda-forge jarvis-tools`"
        )

    # Load all of the 3D data from JARVIS. This gives us a list of dictionaries
    # TODO: In the future, we can include other datasets like their 2D dataset.
    data = jarvis_helper("dft_3d")

    # Now iterate through all the data -- which is a list of dictionaries.
    # We convert the data into a pymatgen object and sanitize it before saving
    # to the Simmate database
    for entry in tqdm(data):

        # The structure is in the atoms field as a dictionary. We pull this data
        # out and convert it to a pymatgen Structure object
        structure = Structure(
            lattice=entry["atoms"]["lattice_mat"],
            species=entry["atoms"]["elements"],
            coords=entry["atoms"]["coords"],
            coords_are_cartesian=entry["atoms"]["cartesian"],
        )

        # now convert the entry to a database object
        structure_db = JarvisStructure.from_toolkit(
            id=entry["jid"].lower(),
            structure=structure,
            energy_above_hull=entry["ehull"] if entry["ehull"] != "na" else None,
        )

        # and save it to our database!
        structure_db.save()



# -*- coding: utf-8 -*-

import os

import pytest

from simmate.conftest import copy_test_files, make_dummy_files
from simmate.calculators.vasp.inputs import Potcar
from simmate.calculators.vasp.tasks.base import VaspTask
from simmate.calculators.vasp.inputs.potcar_mappings import PBE_ELEMENT_MAPPINGS


class DummyTask(VaspTask):
    """
    A minimal example VaspTask that is just for testing
    """

    functional = "PBE"
    potcar_mappings = PBE_ELEMENT_MAPPINGS
    confirm_convergence = True
    pre_sanitize_structure = True
    incar = dict(
        PREC="Low",
        EDIFF__per_atom=2e-3,  # to ensure structure-specific kwargs
        KSPACING=0.75,
    )


def test_base_setup(structure, tmpdir, mocker):

    # estabilish filenames that we make and commonly reference
    incar_filename = os.path.join(tmpdir, "INCAR")
    poscar_filename = os.path.join(tmpdir, "POSCAR")
    potcar_filename = os.path.join(tmpdir, "POTCAR")

    # Because we won't have POTCARs accessible, we need to cover this function
    # call -- specifically have it pretend to make a file
    mocker.patch.object(
        Potcar,
        "to_file_from_type",
        return_value=make_dummy_files(potcar_filename),
    )

    # try to make input files in the tmpdir
    DummyTask.setup(directory=tmpdir, structure=structure)
    assert os.path.exists(incar_filename)
    assert os.path.exists(poscar_filename)
    assert os.path.exists(potcar_filename)
    Potcar.to_file_from_type.assert_called_with(
        structure.composition.elements,
        "PBE",
        potcar_filename,
        PBE_ELEMENT_MAPPINGS,
    )


def test_base_workup(tmpdir):
    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="base",
    )

    # estabilish filenames that we make and commonly reference
    summary_filename = os.path.join(tmpdir, "simmate_summary.yaml")
    vasprun_filename = os.path.join(tmpdir, "vasprun.xml")

    # run the full workup
    DummyTask.workup(tmpdir)
    assert os.path.exists(summary_filename)

    # run the workup again with a malformed xml
    with open(vasprun_filename, "r") as file:
        contents = file.readlines()
    with open(vasprun_filename, "w") as file:
        file.writelines(contents[50])
    with pytest.raises(Exception):
        DummyTask.workup(tmpdir)



# -*- coding: utf-8 -*-

import os

from simmate.toolkit import Structure
from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class LargeSigma(ErrorHandler):
    """
    When ISMEAR >= 0 (Gaussian or Methfessel-Paxton), we need to monitor the
    magnitude of the entropy term T*S. If the entropy term is larger than
    1 meV/atom, then we reduce value of SIGMA. See VASP documentation for ISMEAR.
    """

    is_monitor = True

    def __init__(self, entropy_per_atom_threshold=0.001):
        self.entropy_per_atom_threshold = entropy_per_atom_threshold

    def check(self, directory: str) -> bool:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # this error is only relevent if we have ISMEAR >= 0. So we return that
        # there is no error otherwise.
        if incar.get("ISMEAR", -1) < 0:
            return False

        # We check for this error in the OUTCAR
        # TODO: I don't have an OUTCAR class written yet so I just read the
        # raw text and search for the entropy information directly
        outcar_filename = os.path.join(directory, "OUTCAR")

        # check to see that the file is there first
        if os.path.exists(outcar_filename):

            # read the file content and then close it
            with open(outcar_filename) as file:
                outcar_lines = file.readlines()

            # also load the structure so we know how many sites there are
            poscar_filename = os.path.join(directory, "POSCAR")
            structure = Structure.from_file(poscar_filename)
            nsites = structure.num_sites

            # iterate through all the lines and look for the entropy value.
            # these lines look like this:
            #   entropy T*S    EENTRO =         0.00000000
            # where the entropy is the final value. If any entropy value
            # exceeds our threshold then we have an error!
            for line in outcar_lines:
                if "entropy T*S" in line:
                    entropy = float(line.split()[-1])
                    entropy_per_atom = entropy / nsites
                    if entropy_per_atom > self.entropy_per_atom_threshold:
                        return True

        # if the file doesn't exist OR the threshold is never hit, the we are
        # not seeing any error.
        return False

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # grab the sigma value being used where the default is 0.2
        current_sigma = incar.get("SIGMA", 0.2)

        # If the sigma value is above 0.08, we try reducing it by 0.06.
        # Note the .00001 added on is to handle rounding issues
        if current_sigma > 0.08001:
            new_sigma = current_sigma - 0.06
            incar["SIGMA"] = new_sigma
            incar.to_file(incar_filename)
            return f"reduced SIGMA from {current_sigma} to {new_sigma}"

        # check what the current KSPACING is. If it's not set, that means it's using
        # the default which is 0.5.
        current_kspacing = incar.get("KSPACING", 0.5)

        # If the sigma value is at or below 0.08 and we still have this error,
        # we may not have a dense enough kpt mesh. We then try decreasing KSPACING
        # by 20% in each direction. which approximately doubles the number
        # of kpoints. We stop trying this if we are below a K-pt density of 0.3
        if current_kspacing > 0.3:
            new_kspacing = current_kspacing * 0.8
            incar["KSPACING"] = new_kspacing
            # rewrite the INCAR with new settings
            incar.to_file(incar_filename)
            return f"switched KSPACING from {current_kspacing} to {new_kspacing}"

        # otherwise we can't do anything else and should flag the calculation
        # as failed
        raise Exception("Unable to fix LargeSigma error")



# -*- coding: utf-8 -*-

import os
import json

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class SubspaceMatrix(ErrorHandler):
    """
    This handler fixes when the sub-space-matrix is not hermitian. On the first
    attempt, we switch evaluation of projection operators to reciprocal space.
    If that doesn't work, the calculation precision is switched to "Accurate".
    """

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["WARNING: Sub-Space-Matrix is not hermitian in DAV"]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # load the error-count file if it exists
        error_count_filename = os.path.join(directory, "simmate_error_counts.json")
        if os.path.exists(error_count_filename):
            with open(error_count_filename) as error_count_file:
                error_counts = json.load(error_count_file)
        # otherwise we are starting with an empty dictionary
        else:
            error_counts = {}

        # The fix is based on the number of times we've already tried to
        # fix this. So let's make sure it's in our error_count dictionary.
        # If it isn't there yet, set the count to 0.
        # Also increase the count by 1, as this is our new addition
        error_counts["subspacematrix"] = error_counts.get("subspacematrix", 0) + 1

        # On our first try, change evalutating projection operators
        # in reciprocal space.
        if error_counts["subspacematrix"] == 0:
            incar["LREAL"] = False
            correction = "set LREAL to False"
        # As a last resort, try changing the precision
        else:
            incar["PREC"] = "Accurate"
            correction = "set PREC to Accurate"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        # rewrite the new error count file
        with open(error_count_filename, "w") as file:
            json.dump(error_counts, file)

        # now return the correction made for logging
        return correction



# -*- coding: utf-8 -*-

from simmate.utilities import get_doc_from_readme

__doc__ = get_doc_from_readme(__file__)

from .atomic_permutation import AtomicPermutation
from .coordinate_perturbation import CoordinatePerturbation
from .heredity_mutation import Heredity
from .lattice_strain import LatticeStrain
from .mirror_mutation import MirrorMutation
from .rotational_mutation import RotationalMutation
from .soft_mutation import SoftMutation



# -*- coding: utf-8 -*-

import os

from pymatgen.io.vasp.inputs import Kpoints
from pymatgen.symmetry.bandstructure import HighSymmKpath
from pymatgen.electronic_structure.plotter import BSPlotter

from simmate.calculators.vasp.inputs import Incar, Poscar, Potcar
from simmate.calculators.vasp.tasks.static_energy.materials_project import (
    MatprojStaticEnergy,
)


class VaspBandStructure(MatprojStaticEnergy):
    """
    A base class for band structure calculations. This is not meant
    to be used directly but instead should be inherited from.

    Your stucture will be converted to the standardized-primitive unitcell so
    that the high-symmetry K-point path can be used.

    This is also a non self-consistent field (non-SCF) calculation and thus uses
    the a fixed charge density from a previous static energy calculation.
    """

    required_files = MatprojStaticEnergy.required_files + ["CHGCAR"]

    # set the KptGrid or KptPath object
    # TODO: in the future, all functionality of this class will be available
    # by giving a KptPath class here.
    kpoints = None
    kpoints_line_density = 20
    """
    Density of k-points to use along high-symmetry lines
    """

    @classmethod
    def setup(cls, directory, structure, **kwargs):
        """
        Writes input files for this calculation. This differs from the normal
        VaspTask setup because it converts the structure to the standard primative
        first and then writes a KPOINT file with using a highsym path.
        """

        # run cleaning and standardizing on structure (based on class attributes)
        structure_cleaned = cls._get_clean_structure(structure, **kwargs)

        # write the poscar file
        Poscar.to_file(structure_cleaned, os.path.join(directory, "POSCAR"))

        # Combine our base incar settings with those of our parallelization settings
        # and then write the incar file
        incar = Incar(**cls.incar) + Incar(**cls.incar_parallel_settings)
        incar.to_file(
            filename=os.path.join(directory, "INCAR"),
            structure=structure_cleaned,
        )

        ##############
        # we need to find the high-symmetry Kpt path. Note that all of this
        # functionality will be moved to the KptPath class and then extended to
        # vasp.inputs.kpoints class. Until those classes are ready, we just use
        # pymatgen here.
        sym_prec = cls.incar.get("SYMPREC", 1e-5) if cls.incar else 1e-5
        kpath = HighSymmKpath(structure_cleaned, symprec=sym_prec)
        frac_k_points, k_points_labels = kpath.get_kpoints(
            line_density=cls.kpoints_line_density,
            coords_are_cartesian=False,
        )
        kpoints = Kpoints(
            comment="Non SCF run along symmetry lines",
            style=Kpoints.supported_modes.Reciprocal,
            num_kpts=len(frac_k_points),
            kpts=frac_k_points,
            labels=k_points_labels,
            kpts_weights=[1] * len(frac_k_points),
        )
        kpoints.write_file(os.path.join(directory, "KPOINTS"))
        ##############

        # write the POTCAR file
        Potcar.to_file_from_type(
            structure_cleaned.composition.elements,
            cls.functional,
            os.path.join(directory, "POTCAR"),
            cls.potcar_mappings,
        )

    @staticmethod
    def _write_output_summary(directory, vasprun):
        """
        In addition to writing the normal VASP output summary, this also plots
        the bandstructure to "band_structure.png"
        """

        # run the normal output
        MatprojStaticEnergy._write_output_summary(directory, vasprun)

        bs_plotter = BSPlotter(vasprun.get_band_structure(line_mode=True))
        plot = bs_plotter.get_plot()
        plot_filename = os.path.join(directory, "band_structure.png")
        plot.savefig(plot_filename)



# -*- coding: utf-8 -*-

import os

import pytest

from pandas import DataFrame

# from pymatgen.io.vasp.outputs import Vasprun

from simmate.toolkit import Structure
from simmate.website.test_app.models import TestStaticEnergy


@pytest.mark.django_db
def test_static_energy_table(structure, tmpdir):

    # test writing columns
    TestStaticEnergy.show_columns()

    # test writing to database
    structure_db = TestStaticEnergy.from_prefect_id(
        prefect_flow_run_id="example-id-123",
        structure=structure,
    )
    structure_db.save()

    # try grabbing the calculation again and make sure it loaded from the
    # database rather than creating a new entry
    structure_db2 = TestStaticEnergy.from_prefect_id(
        prefect_flow_run_id="example-id-123",
    )
    assert structure_db.id == structure_db2.id

    # test converting back to toolkit and ensuring the structure is the
    # same as before
    structure_new = structure_db.to_toolkit()
    assert structure == structure_new

    # test converting search results to dataframe and to toolkit
    df = TestStaticEnergy.objects.to_dataframe()
    assert isinstance(df, DataFrame)
    structures = TestStaticEnergy.objects.to_toolkit()
    assert isinstance(structures, list)
    assert isinstance(structures[0], Structure)

    # update the database entry using a Vasprun result.
    # TODO:

    # test writing and reloading these from and archive
    archive_filename = os.path.join(tmpdir, "archive.zip")
    TestStaticEnergy.objects.to_archive(archive_filename)
    TestStaticEnergy.load_archive(
        archive_filename,
        confirm_override=True,
        delete_on_completion=True,
    )



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.band_structure import VaspBandStructure
from simmate.calculators.vasp.tasks.relaxation import MatprojRelaxation


class MatprojBandStructure(VaspBandStructure, MatprojRelaxation):
    """
    This task is a reimplementation of pymatgen's
    [MPNonSCFSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPNonSCFSet)
    with mode="line".

    Calculates the band structure using Materials Project HSE settings.
    """

    incar = MatprojRelaxation.incar.copy()
    incar.update(
        IBRION=-1,
        LCHARG=False,
        LORBIT=11,
        LWAVE=False,
        NSW=0,
        ISYM=0,
        ICHARG=11,
        ISMEAR=0,
        SIGMA=0.01,
    )
    incar.pop("MAGMOM__smart_magmom")



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.relaxation.mit import MITRelaxation

from simmate.calculators.vasp.error_handlers import (
    TetrahedronMesh,
    Eddrmm,
    NonConverging,
)


class MITStaticEnergy(MITRelaxation):
    """
    Runs a VASP static energy calculation using MIT Project settings.

    This is identical to relaxation/mit, but just a single ionic step.
    """

    # The settings used for this calculation are based on the MITRelaxation, but
    # we are updating/adding new settings here.
    incar = MITRelaxation.incar.copy()
    incar.update(
        dict(
            ALGO="Normal",
            IBRION=-1,  # (optional) locks everything between ionic steps
            NSW=0,  # this is the main static energy setting
            KSPACING=0.4,  # was 0.5 # !!! This is where we are different from pymatgen right now
        )
    )



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class Edwav(ErrorHandler):
    """
    When the gradient is not orthogonal, we make a fix depending on what IMSEAR
    is set to.
    """

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["EDWAV: internal error, the gradient is not orthogonal"]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # We make a fix only if IMSEAR is set less than 0
        if incar.get("ISMEAR", 1) < 0:
            incar["ISMEAR"] = "0"
            incar["SIGMA"] = 0.05
            correction = "switched SIGMA to 0.05 and ALGO to Normal"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

import pytest
from pytest_django.asserts import assertTemplateUsed
from django.urls import reverse

from simmate.workflow_engine import Workflow
from simmate.workflows.utilities import (
    get_workflow_types,
    get_list_of_all_workflows,
    get_workflow,
)

ALL_WORKFLOWS = get_list_of_all_workflows()


def test_workflows_view(client):
    response = client.get("/workflows/")
    assert response.status_code == 200
    assertTemplateUsed(response, "workflows/all.html")


@pytest.mark.parametrize("workflow_type", get_workflow_types())
def test_workflows_by_type_view(client, workflow_type):

    # grabs f"/workflows/{workflow_type}/"
    url = reverse(
        "workflows_by_type",
        kwargs={"workflow_type": workflow_type},
    )

    response = client.get(url)
    assert response.status_code == 200
    assertTemplateUsed(response, "workflows/by_type.html")


@pytest.mark.django_db
@pytest.mark.parametrize("workflow_name", ALL_WORKFLOWS)
def test_workflow_detail_view(client, workflow_name):

    # BUG: I assume .vasp. in the view for now
    if workflow_name == "restart.simmate.automatic":
        return

    workflow = get_workflow(workflow_name)

    # list view
    # grabs f"/workflows/{type}/{name}/")
    url = reverse(
        "workflow_detail",
        kwargs={
            "workflow_type": workflow.name_project,
            "workflow_name": workflow.name_preset,
        },
    )

    response = client.get(url)
    assert response.status_code == 200
    assertTemplateUsed(response, "workflows/detail.html")

    # detail view - found
    # grabs... "/workflows/static-energy/mit/1/"
    # TODO: how should I populate with test data?

    # detail view - not found
    # grabs... "/workflows/static-energy/mit/999/"
    url = reverse(
        "workflow_run_detail",
        kwargs={
            "workflow_type": workflow.name_project,
            "workflow_name": workflow.name_preset,
            "pk": 999,
        },
    )

    response = client.get(url)
    assert response.status_code == 404
    # assertTemplateUsed(response, "workflows/detail_run.html")


# @pytest.mark.parametrize("workflow_name", ALL_WORKFLOWS) # TODO
@pytest.mark.django_db
def test_workflow_submit_view(client, sample_structures, mocker):

    client.login(username="test_user", password="test_password")

    url = reverse(
        "workflow_submit",
        kwargs={
            "workflow_type": "static-energy",
            "workflow_name": "mit",
        },
    )

    # loading blank page
    response = client.get(url)
    assert response.status_code == 200
    assertTemplateUsed(response, "workflows/submit.html")

    # grab a test structure to submit
    structure = sample_structures["C_mp-48_primitive"]

    # I don't want to actually run the workflow, so I override the run method
    mocker.patch.object(
        Workflow,
        "run_cloud",
        return_value="example-run-id-1234",
    )

    # submit the form
    response = client.post(
        url,
        {
            "labels": "test_label1",
            # This is just JSON for a NaCl structure
            "structure_json": structure.to_json(),
        },
    )

    assert (
        response.url == "https://cloud.prefect.io/simmate/flow-run/example-run-id-1234"
    )
    Workflow.run_cloud.assert_called_with(
        structure=structure,
        labels=["test_label1"],
        source=None,
        copy_previous_directory=False,
        wait_for_run=False,
        pre_sanitize_structure=False,
        pre_standardize_structure=False,
        # parameters not deserialized yet so these will still be present
        command="",
        directory="",
        is_restart=False,
    )



# -*- coding: utf-8 -*-

from django.http import HttpRequest

from rest_framework.viewsets import GenericViewSet

# from rest_framework.generics import GenericAPIView
from rest_framework.response import Response
from rest_framework.serializers import ModelSerializer

from simmate.website.core_components.filters import DatabaseTableFilter
from simmate.database.base_data_types import DatabaseTable


class SimmateAPIViewSet(GenericViewSet):
    """
    Example use:

    ``` python
    # in views.py
    class ExampleTableViewSet(SimmateAPIViewSet):
        table = ExampleTable
        template_list = "custom_app/list.html"
        template_retrieve = "custom_app/retrieve.html"

    # in urls.py
    path(
        route="example-table",
        view=ExampleTableViewSet.list_view,
        name="example-list",
    )
    path(
        route="example-table/<int:id>",
        view=ExampleTableViewSet.retrieve_view,
        name="example-retrieve",
    )
    ```
    """

    table: DatabaseTable = None

    template_list: str = None

    template_retrieve: str = None

    def get_list_response(self, request: HttpRequest, *args, **kwargs) -> Response:

        # This code is modified from the ListModelMixin, where instead of returning
        # a response, we perform additional introspection first.
        # https://github.com/encode/django-rest-framework/blob/master/rest_framework/mixins.py#L33

        # self.format_kwarg --> not sure why this always returns None, so I
        # grab the format from the request instead. If it isn't listed, then
        # I'm using the default which is html.
        self._format_kwarg = request.GET.get("format", "html")

        queryset = self.filter_queryset(self.get_queryset())

        # attempt to paginate the results
        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
        else:
            serializer = self.get_serializer(queryset, many=True)

        # If don't have the html format, we follow simple logic from the
        # original ListModelMixin method
        if self._format_kwarg != "html":
            if page is not None:
                return self.get_paginated_response(serializer.data)
            else:
                return Response(serializer.data)

        # otherwise we assume the html format.
        else:
            filterset = self.filterset_class(request.GET)
            data = {
                # "filterset": filterset, # not used at the momemnt
                "filterset_mixins": filterset.get_mixin_names(),
                "form": filterset.form,
                "extra_filters": filterset.get_extra_filters(),
                "calculations": serializer.instance,  # return python objs, not dict
                "ncalculations_matching": queryset.count(),
                "ncalculations_possible": self.get_queryset().count(),
                **self.paginator.get_html_context(),
                **self.get_list_context(request, **kwargs),
            }
            return Response(data)

    def get_retrieve_response(self, request: HttpRequest, *args, **kwargs) -> Response:

        # self.format_kwarg --> not sure why this always returns None, so I
        # grab the format from the request instead. If it isn't listed, then
        # I'm using the default which is html.
        self._format_kwarg = request.GET.get("format", "html")

        # ---------------------------------------------------
        # This code is from the RetrieveModelMixin, where instead of returning
        # a response, we perform additional introspection first
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        # return Response(serializer.data) <--- removed from original
        # ---------------------------------------------------

        if self._format_kwarg == "html":
            # we want to return a python object, not a serialized dictionary
            # because the object has more attributes and methods attached to it.
            calculation = serializer.instance
            data = {
                "calculation": calculation,
                "table_mixins": calculation.get_mixin_names(),
                "extra_columns": calculation.get_extra_columns(),
                **self.get_retrieve_context(request, **kwargs),
            }
            return Response(data)
        else:
            return Response(serializer.data)

    @classmethod
    def from_table(
        cls,
        table: DatabaseTable,
        view_type: str,
        **kwargs,
    ):
        # For all tables, we share all the data -- no columns are hidden. Therefore
        # the code for the Serializer is always the same.
        class NewSerializer(ModelSerializer):
            class Meta:
                model = table
                fields = "__all__"

        # Build the filter from our base class using default settings
        NewFilterSet = DatabaseTableFilter.from_table(table)

        # For the source dataset, not all tables have a "created_at" column, but
        # when they do, we want to return results with the most recent additions first
        # by default. Ordering can also be overwritten by passing "ordering=..."
        # to the URL.
        default_ordering_field = (
            "-created_at" if hasattr(table, "created_at") else table._meta.pk.name
        )

        # we also want to preload spacegroup for the structure mixin
        intial_queryset = table.objects.all()
        if hasattr(table, "spacegroup"):
            intial_queryset = intial_queryset.select_related("spacegroup")

        NewViewSet = type(
            f"{table.__name__}ViewSet",
            (cls,),
            dict(
                queryset=intial_queryset,
                serializer_class=NewSerializer,
                filterset_class=NewFilterSet,
                ordering_fields="__all__",  # allowed to order by any field
                ordering=[default_ordering_field],  # set default order
                **kwargs,
            ),
        )

        # these two parameters depend on the view_type
        if view_type == "list":
            NewViewSet.template_name = cls.template_list
            return NewViewSet.as_view({"get": "get_list_response"})
        elif view_type == "retrieve":
            NewViewSet.template_name = cls.template_retrieve
            return NewViewSet.as_view({"get": "get_retrieve_response"})
        else:
            raise Exception("Unknown view type. Must be 'list' or 'retrieve'.")

    # -------------------------------------------------------------------------

    # METHODS FOR STATIC VIEWS

    @classmethod
    def list_view(cls, request, **request_kwargs):
        view = cls.from_table(table=cls.table, view_type="list")
        return view(request, **request_kwargs)

    @classmethod
    def retrieve_view(cls, request, **request_kwargs):
        view = cls.from_table(table=cls.table, view_type="retrieve")
        return view(request, **request_kwargs)

    # -------------------------------------------------------------------------

    # METHODS FOR DYNAMIC VIEWS

    # NOTE: These dynamically create a serializer and a view EVERY TIME a
    # URL is requested. This means...
    #   1. there is no pre-set api that exists. The exisiting api must be inferred
    #       from lower level workflows and their tables
    #   2. these views will be very inefficient if queried by a script
    # I chose dynamic creation over creating all endpoints on-startup to prevent
    # the `from simmate.database import connect` method from taking too long --
    # as that would require import all workflows on start-up. However, if the
    # (1) api spec or (2) speed of this method ever becomes an issue, I can
    # address these by either...
    #   1. having a utility that prints out the full API spec but isn't called on startup
    #   2. making all APIViews up-

    def get_table(self, request: HttpRequest, *args, **kwargs) -> Response:
        if not self.table:
            raise NotImplementedError()
        else:
            return self.table

    @classmethod
    def dynamic_list_view(cls, request, **request_kwargs):
        table = cls.get_table(request, **request_kwargs)
        view = cls.from_table(table=table, view_type="list")
        return view(request, **request_kwargs)

    @classmethod
    def dynamic_retrieve_view(cls, request, **request_kwargs):
        table = cls.get_table(request, **request_kwargs)
        view = cls.from_table(table=table, view_type="retrieve")
        return view(request, **request_kwargs)

    # -------------------------------------------------------------------------

    # OPTIONAL METHODS FOR SUPPLYING EXTRA CONTEXT TO HTML TEMPLATES

    def get_list_context(self, request, **request_kwargs) -> dict:
        """
        This defines extra context that should be passed to the template when
        using format=html when a list-view is requested. By default, no extra
        context is returned.
        """
        return {}

    def get_retrieve_context(self, request, **request_kwargs) -> dict:
        """
        This defines extra context that should be passed to the template when
        using format=html when a retrieve-view is requested. By default, no extra
        context is returned.
        """
        return {}



# -*- coding: utf-8 -*-

import pytest
from pytest_django.asserts import assertTemplateUsed


@pytest.mark.django_db
def test_home_view(client):

    # test initial view
    response = client.get("/")
    assert response.status_code == 200
    assertTemplateUsed(response, "home/home.html")

    # test submission of the form
    response = client.post(
        "/",
        {
            "chemical_system": "Y-F-C",
            "materials_project": True,
            "jarvis": True,
        },
    )
    assert response.status_code == 200


def test_extras_view(client):
    response = client.get("/extras/")
    assert response.status_code == 200
    assertTemplateUsed(response, "core_components/extras.html")


def test_contact_view(client):
    response = client.get("/contact/")
    assert response.status_code == 200
    assertTemplateUsed(response, "core_components/contact.html")


def test_about_view(client):
    response = client.get("/about/")
    assert response.status_code == 200
    assertTemplateUsed(response, "core_components/about.html")



# -*- coding: utf-8 -*-

from typing import List

from simmate.toolkit import Composition, Structure
from simmate.database import connect
from simmate.database.third_parties import (
    AflowStructure,
    CodStructure,
    JarvisStructure,
    MatprojStructure,
    OqmdStructure,
)


def get_known_structures(
    composition: Composition,
    strict_nsites: bool = False,
) -> List[Structure]:
    """
    Goes through all database tables in the `simmate.database.third_parties`
    module and grabs all structures with a matching composition.

    Each database table must be populated -- otherwise this function will
    return an empty list.
    """

    databases_to_search = [
        AflowStructure,
        CodStructure,
        JarvisStructure,
        MatprojStructure,
        OqmdStructure,
    ]

    structures = []

    for database_table in databases_to_search:

        # If there is a limit to nsites, we want to match the composition formula
        # exactly. Otherwise we can match the reduced formula.
        if strict_nsites:
            search_results = database_table.objects.filter(
                formula_full=composition.formula,
            )
        else:
            search_results = database_table.objects.filter(
                formula_reduced=composition.reduced_formula,
            )

        # Convert to toolkit structures so that users can run analyses
        structures += search_results.to_toolkit()

    return structures



# -*- coding: utf-8 -*-

from pymatgen.analysis.diffusion.neb.pathfinder import (
    MigrationHop as PymatgenMigrationHop,
)


class MigrationHop(PymatgenMigrationHop):
    def as_dict(self):
        # Pymatgen.diffusion has a broken as_dict method, so I need to fix this
        # here. I don't want to raise an error because I don't want issues
        # with the higher-level serialize_parameters utility
        return (
            "MigrationHop is not MSONable, and this bug is within pymatgen diffusion."
            "See https://github.com/materialsvirtuallab/pymatgen-analysis-diffusion/pull/264"
        )

    @classmethod
    def from_dynamic(cls, migration_hop):
        """
        This is an experimental feature. The code here is a repurposing of
        Structre.from_dynamic so consider making a general class for
        from_dynamic methods.
        """
        is_from_past_calc = False

        if isinstance(migration_hop, PymatgenMigrationHop):
            migration_hop_cleaned = migration_hop
        elif isinstance(migration_hop, dict) and "@module" in migration_hop.keys():
            migration_hop_cleaned = cls.from_dict(migration_hop)
        elif (
            isinstance(migration_hop, dict)
            and "migration_hop_table" in migration_hop.keys()
        ):
            is_from_past_calc = True
            migration_hop_cleaned = cls.from_database_dict(migration_hop)
        else:
            raise Exception("Unknown format provided for migration_hop input.")
        migration_hop_cleaned.is_from_past_calc = is_from_past_calc

        return migration_hop_cleaned

    @classmethod
    def from_database_dict(cls, migration_hop: dict):
        """
        This is an experimental feature. The code here is a repurposing of
        Structre.from_dynamic so consider making a general class for
        from_dynamic methods.

        ex:
            {
                "migration_hop_table": "MITMigrationHop",
                "migration_hop_id": 1,
            }
        """

        # Imports are done locally to keep this class modular.
        from simmate.database import connect
        from simmate.website.workflows import models as all_datatables
        from django.utils.module_loading import import_string

        datatable_str = migration_hop["migration_hop_table"]

        if hasattr(all_datatables, datatable_str):
            datatable = getattr(all_datatables, datatable_str)
        else:
            datatable = import_string(datatable_str)
        # for now I only support migration_hop_id
        migration_hop_db = datatable.objects.get(id=migration_hop["migration_hop_id"])
        migration_hop_cleaned = migration_hop_db.to_toolkit()
        migration_hop_cleaned.database_entry = migration_hop_db

        return migration_hop_cleaned



# -*- coding: utf-8 -*-

from numpy.random import normal as numpy_random_normal


class NormallyDistributedVectors:

    # This class creates random coordinates (x,y,z) that follow a normal (Guassian)
    # distribution. The random values will between min_value and max_value, and also
    # extra conditions can be added. For example, ['x>=y', 'x<z*2', 'y<z+1']

    def __init__(
        self, min_value=0, max_value=1, extra_conditions=[], center=None, standdev=None
    ):

        # Extra conditions should be a list of strings that use x,y,z for vector
        # positions. For example, ['x>=y', 'x<z*2', 'y<z+1']

        self.min_value = min_value
        self.max_value = max_value
        self.extra_conditions = extra_conditions

        # Set the center of the normal (Guassian) distribution
        # If a value is provided we use it. Otherwise we set it to the middle
        # of the min/max values
        self.center = center if center else (max_value + min_value) / 2

        # Set the standard deviation of the normal (Guassian) distribution
        # If a value is provided we use it. Otherwise we set it to 1/5*(max-min)
        self.standdev = standdev if standdev else (max_value - min_value) / 5

    def new_vector(self):

        # We want to attempt to make a vector until it meets all of the conditions
        # that we set above. We therefore have a while loop. To start the whileloop,
        # we need this conditon set to false
        is_valid_vector = False
        while not is_valid_vector:

            # generate random 3x1 matrix where values are normally distributed
            vector = numpy_random_normal(loc=self.center, scale=self.standdev, size=3)

            # check if all values are between min/max values specified
            is_valid_vector = (vector.min() >= self.min_value) and (
                vector.max() <= self.max_value
            )
            # if the check is False, use 'continue' to restart the while loop
            if not is_valid_vector:
                continue

            # check that all extra conditions are met and assume it is valid
            # until proven otherwise.
            for condition in self.extra_conditions:
                x, y, z = vector
                is_valid_vector = eval(condition, None, dict(x=x, y=y, z=z))
                # If this condition isn't met (check=False), then we can exit
                # the loop and try with a new vector
                if not is_valid_vector:
                    break

            # If the for-loop above 'breaks' at any point, check will be False
            # and the while-loop will restart. If the for-loop goes through all
            # extra_conditions and never breaks, check will be True and the
            # while-loop will finish

        return vector



# -*- coding: utf-8 -*-

"""
This converter simply imports a pymatgen class. With this you can quickly
convert between Simmate.toolkit and ASE.atoms objects.

``` python

from simmate.toolkit import Structure
from simmate.file_converters.structure.ase import AseAtomsAdaptor


initial_structure = Structure.from_file("example.cif")

# convert to ASE
new_ase_object = AseAtomsAdaptor.get_atoms(structure)

# convert back to PyMatgen/Simmate
new_toolkit_object = AseAtomsAdaptor.get_structure(new_ase_object)

# alternatively, you can convert back to PyMatgen Molecule
new_molecule_object = AseAtomsAdaptor.get_molecule(new_ase_object)
```
"""


from pymatgen.io.ase import AseAtomsAdaptor

__all__ = ["AseAtomsAdaptor"]



# -*- coding: utf-8 -*-

import pytest

from pandas import DataFrame

# from pymatgen.io.vasp.outputs import Vasprun

from simmate.toolkit import Structure
from simmate.website.test_app.models import TestRelaxation, TestIonicStep


@pytest.mark.django_db
def test_static_energy_table(structure):

    # test writing columns
    TestRelaxation.show_columns()
    TestIonicStep.show_columns()

    # test writing to database
    structure_db = TestRelaxation.from_prefect_id(
        prefect_flow_run_id="example-id-123",
        structure=structure,
    )
    structure_db.save()

    # try grabbing the calculation again and make sure it loaded from the
    # database rather than creating a new entry
    structure_db2 = TestRelaxation.from_prefect_id(prefect_flow_run_id="example-id-123")
    assert structure_db.id == structure_db2.id

    # test converting back to toolkit and ensuring the structure is the
    # same as before
    structure_new = structure_db.to_toolkit()
    assert structure == structure_new

    # test converting search results to dataframe and to toolkit
    df = TestRelaxation.objects.to_dataframe()
    assert isinstance(df, DataFrame)
    structures = TestRelaxation.objects.to_toolkit()
    assert isinstance(structures, list)
    assert isinstance(structures[0], Structure)



# -*- coding: utf-8 -*-

"""
This defines commands for managing your Simmate database. All commands are 
accessible through the `simmate database` command.
"""

import click


@click.group()
def database():
    """A group of commands for managing your database."""
    pass


@database.command()
@click.option(
    "--confirm-delete",
    is_flag=True,
    help="automatically confirms that you want to delete your existing database",
)
@click.option(
    "--use-prebuilt",
    default=None,
    type=bool,
    help="automatically says yes/no a prebuilt database (only applies if using sqlite)",
)
def reset(confirm_delete, use_prebuilt):
    """Removes any existing data and sets up a clean database."""

    # make sure the user knows what they are doing and actually wants to continue
    if not confirm_delete:
        click.confirm(
            "Note that this deletes your current database and cannot be undone. "
            "We highly recommend you make a copy of your database before doing this. "
            "Do you still want to continue?",
            abort=True,
        )

    # if the user has a SQLite3 backend, ask if they'd like to use a prebuilt
    # database to begin
    from simmate.configuration.django.settings import DATABASES

    using_sqlite = DATABASES["default"]["ENGINE"] == "django.db.backends.sqlite3"
    if using_sqlite and use_prebuilt == None:
        use_prebuilt = click.confirm(
            "It looks like you are using the default database backend (sqlite3). "
            "Would you like to use a prebuilt-database with all third-party data "
            "already loaded? If this is the first time you using the prebuild, "
            "this will involve a ~3GB download and will unpack to roughly 7GB. A "
            "backup of the download will be stored as well (so another ~3GB disk "
            "space will be used). We recommend answering 'yes' for beginners."
        )

    # We can now proceed with reseting the database
    click.echo("Removing database and rebuilding...")

    from simmate.database import connect
    from simmate.database.utilities import reset_database

    reset_database(use_prebuilt=use_prebuilt)

    # Let the user know everything succeeded
    click.echo("Success! Your database has been reset.")


@database.command()
def update():
    """Updates the database with any changes made."""

    # check Django if there are any updates to be made
    click.echo("Checking for updates...")

    from simmate.database import connect
    from simmate.database.utilities import update_database

    update_database()

    # Let the user know everything succeeded
    click.echo("Success! Your database tables are now up to date.")


@database.command()
@click.option(
    "--filename",
    "-f",
    default="database_dump.json",
    help="the file to write the all the JSON data to",
)
def dumpdata(filename):
    """Takes the Simmate database and writes it to a json file."""

    # Begin writing the database to the json file.
    click.echo("Writing all data to JSON...")

    from simmate.database import connect
    from simmate.database.utilities import dump_database_to_json

    dump_database_to_json(filename=filename)

    # Let the user know everything succeeded
    click.echo(
        f"Success! You should now see the file {filename} with all of your data."
    )


@database.command()
@click.option(
    "--filename",
    "-f",
    default="database_dump.json",
    help="the file to load the all the JSON data from",
)
def loaddata(filename):
    """Takes a JSON database and loads it into the Simmate database."""

    # Begin writing the database to the json file.
    click.echo("Loading all data from JSON...")

    from simmate.database import connect
    from simmate.database.utilities import load_database_from_json

    load_database_from_json(filename=filename)

    # Let the user know everything succeeded
    click.echo(
        f"Success! You now have all the data from {filename} available in your database."
    )


@database.command()
@click.option(
    "--parallel",
    is_flag=True,
    help="(not allowed for SQLite3) loads data using parallel processing",
)
def load_remote_archives(parallel):
    """
    Goes through all third-party databases and loads their most recent remote
    archives (if available). This utility helps with initializing a new
    database build.

    WARNING:
    This can take several hours to run and there is no pause/continuation
    implemented. This runs substantially faster when you are using a cloud
    database backend (e.g. Postgres) and use the `--parallel` flag.

    If you are using SQLite, we highly recommend using loading the default build
    when you run `simmate database reset` instead.d.
    """

    from simmate.database import connect
    from simmate.database.third_parties import load_remote_archives

    load_remote_archives(parallel=parallel)
    click.echo("Success! Your database now contains all third-party data.")


# explicitly list functions so that pdoc doesn't skip them
__all__ = [
    "database",
    "reset",
    "update",
    "dumpdata",
    "loaddata",
    "load_remote_archives",
]



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.relaxation import MatprojRelaxation


class MatVirtualLabElastic(MatprojRelaxation):
    """
    This task is a reimplementation of pymatgen's
    [MVLElasticSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MVLElasticSet).
    """

    incar = MatprojRelaxation.incar.copy()
    incar.update(
        dict(
            IBRION=6,
            NFREE=2,
            POTIM=0.015,  # pymatgen uses this as a parameter
        )
    )
    # incar.pop("NPAR")  # pymatgen forcibly removes NPAR



# -*- coding: utf-8 -*-

import os
import json
import itertools
import subprocess

import numpy

from simmate.visualization.structure.blender.configuration import get_blender_command


def make_blender_structure(structure, filename="simmate_structure.blend"):

    # load the base blender command for use in function calls below
    BLENDER_COMMAND = get_blender_command()
    # OPTIMIZE: ideally I would load this outside the function so that it is only
    # loaded once. Here, I read a yaml file repeatedly. There should be a better
    # way to silently catch errors when blender isn't installed.

    # This function simply serializes a pymatgen structure object to json
    # and then calls a blender script (make_structure.py) that uses this data
    # BUG: Make sure strings are dumped using single quotes so that this
    # doesn't conflict with the command line.
    sites = json.dumps(serialize_structure_sites(structure)).replace('"', "'")
    lattice = json.dumps(structure.lattice.matrix.tolist())

    # The location of the make_structure.py
    executable_directory = os.path.dirname(__file__)
    path_to_script = os.path.join(executable_directory, "scripts", "make_structure.py")

    # Now build all of the our serialized structure data and settings together
    # into the blender command that we will call via the command line
    command = (
        f"{BLENDER_COMMAND} --background --factory-startup --python {path_to_script} "
        f'-- --sites="{sites}" --lattice="{lattice}" --save="{filename}"'
    )

    # Now run the command
    result = subprocess.run(
        command,
        shell=True,  # to access commands in the path
        capture_output=True,  # capture any ouput + error logs
    )

    return result


def serialize_structure_sites(structure):
    # NOTE: You only need to call make_blender_structure() as it calls
    # serialize_structure within it.

    # We collect all sites to draw here. Each one is this list will be a tuple
    # of... (element_symbol, radius, cartesian_coordinates)
    # For example, ("Na", 0.75, [0.5, 0.5, 0.5])
    # We also convert the coordinates from numpy to a list here.
    sites_to_draw = []

    # we repeadedly need the lattice matrix below so we just grab it upfront here
    lattice_matrix = structure.lattice.matrix

    # for each site in the structure, we want to gather all site coordinates
    # that we want to display. Note this is more than just structure.cart_coords
    # Because we want symmetrical equivalents that are also on the cell axes.
    # For example, if there a site at (0,0,0) then we also want to display the
    # site at (1,1,1).
    for site in structure:

        # Grab the base info that is the same for all site images here.
        # TODO: We fix the radius for now but may do oxidation analysis for
        # accurate ionic radii in the future
        radius = 0.75
        element = site.specie.symbol

        # first store this base site to our site collection.
        sites_to_draw.append((element, radius, site.coords.tolist()))

        # If a site has a fractional coordinate that is close to zero, then
        # that means we should duplicate the site along that edge of the lattice.
        # For example, (0, 0.5, 0.5) should be duplicated along the a-vector
        # and thus add the site (1, 0.5, 0.5).
        # We do this by checking the fractional coordinates and seeing if each
        # is close to 0.
        zero_elements = [
            i for i, x in enumerate(site.frac_coords) if numpy.isclose(x, 0, atol=0.05)
        ]

        # If more than one value is close to zero, then we need to add multiple
        # duplicate sites. For example, (0,0,0.5) would need us to add
        # (0, 1, 0.5), (1, 0, 0.5), and (1, 1, 0.5). So here we go through
        # and find all permutations that we need to add.
        permutatons = [
            combination
            for n in range(1, len(zero_elements) + 1)
            for combination in itertools.combinations(zero_elements, n)
        ]

        # Now let's iterate through each permutation and add it to our sites list
        for permutaton in permutatons:

            # make the vector that we need to add to the base site. For example,
            # if the permutation is (0,2) then we would do...
            # coords + [1, 0, 1]. Note I use fractional coords in this example
            # but use cartesians coords below.
            shift_vector = numpy.zeros(3)
            for x in permutaton:
                shift_vector = numpy.add(shift_vector, lattice_matrix[x])
            new_coords = site.coords + shift_vector
            sites_to_draw.append((element, radius, new_coords.tolist()))

        # We now repeat the above process but now with sites that have coordinate
        # that are close to 1. The key difference here is that we subract 1 instead
        # of adding 1 like we did above. e.g (1, 0.5, 0.5) becomes (0, 0.5, 0.5)
        zero_elements = [
            i for i, x in enumerate(site.frac_coords) if numpy.isclose(x, 1, atol=0.05)
        ]
        permutatons = [
            combination
            for n in range(1, len(zero_elements) + 1)
            for combination in itertools.combinations(zero_elements, n)
        ]
        for permutaton in permutatons:
            shift_vector = numpy.zeros(3)
            for x in permutaton:
                shift_vector = numpy.subtract(shift_vector, lattice_matrix[x])
            new_coords = site.coords + shift_vector
            sites_to_draw.append((element, radius, new_coords.tolist()))

    ## Add bonded sites outside of unticell
    ### Two issues exist for this code:
    #### 1. doesn't find bonded sites to atoms that were duplicated along edges
    #### 2. duplicate atoms are added to sites_to_draw
    # structure_graph = CrystalNN().get_bonded_structure(structure)
    # for n in range(len(structure_graph.structure)):
    #    connected_sites = structure_graph.get_connected_sites(n)
    #    for connected_site in connected_sites:
    #        if connected_site.jimage != (0, 0, 0):
    #            sites_to_draw.append(connected_site.site.coords)

    return sites_to_draw



# -*- coding: utf-8 -*-

import os
import yaml
from pandas import DataFrame

from pymatgen.io.vasp.outputs import Chgcar
from pymatgen.io.vasp import Potcar

from simmate.toolkit import Structure
from simmate.workflow_engine import S3Task
from simmate.calculators.bader.outputs import ACF
from simmate.calculators.bader.tasks import CombineCHGCARs


class BaderAnalysis(S3Task):

    command = "bader CHGCAR -ref CHGCAR_sum -b weight > bader.out"
    """
    The command to call the executable, which is typically bader. Note we
    use the `-b weight` by default, which means we apply the weight method for
    partitioning from of 
    [Yu and Trinkle](http://theory.cm.utexas.edu/henkelman/code/bader/download/yu11_064111.pdf).
    
    This command is modified to use the `-ref` file as the reference for determining
    zero-flux surfaces when partitioning the CHGCAR. 
    
    There are cases where also use structures that contain "empty atoms" in them.
    This is to help with partitioning of electrides, which possess electron 
    density that is not associated with any atomic orbital. For these cases,
    you will see files like "CHGCAR_empty" used in the command.
    """

    required_files = ["CHGCAR_sum", "POTCAR"]

    @classmethod
    def workup(cls, directory: str):
        """
        A basic workup process that reads Bader analysis results from the ACF.dat
        file and calculates the corresponding oxidation states with the existing
        POTCAR files.
        """

        # load the ACF.dat file
        acf_filename = os.path.join(directory, "ACF.dat")
        dataframe, extra_data = ACF(filename=acf_filename)

        # load the electron counts used by VASP from the POTCAR files
        # OPTIMIZE: this can be much faster if I have a reference file
        potcar_filename = os.path.join(directory, "POTCAR")
        potcars = Potcar.from_file(potcar_filename)
        nelectron_data = {}
        # the result is a list because there can be multiple element potcars
        # in the file (e.g. for NaCl, POTCAR = POTCAR_Na + POTCAR_Cl)
        for potcar in potcars:
            nelectron_data.update({potcar.element: potcar.nelectrons})

        # SPECIAL CASE: in scenarios where empty atoms are added to the structure,
        # we should grab that modified structure instead of the one from the POSCAR.
        chgcar_filename = os.path.join(directory, "CHGCAR")
        chgcar_empty_filename = os.path.join(directory, "CHGCAR_empty")

        # the empty file will always take preference
        if os.path.exists(chgcar_empty_filename):
            chgcar = Chgcar.from_file(chgcar_empty_filename)
            structure = chgcar.structure
            # We typically use hydrogen ("H") as the empty atom, so we will
            # need to add this to our element list for oxidation analysis.
            # We use 0 for electron count because this is an 'empty' atom, and
            # not actually Hydrogen
            nelectron_data.update({"H": 0})
        # otherwise, grab the structure from the CHGCAR
        # OPTIMIZE: consider grabbing from the POSCAR or CONTCAR for speed
        else:
            chgcar = Chgcar.from_file(chgcar_filename)
            structure = chgcar.structure

        # Calculate the oxidation state of each site where it is simply the
        # change in number of electrons associated with it from vasp potcar vs
        # the bader charge I also add the element strings for filtering functionality
        elements = []
        oxi_state_data = []
        for site, site_charge in zip(structure, dataframe.charge.values):
            element_str = site.specie.name
            elements.append(element_str)
            oxi_state = nelectron_data[element_str] - site_charge
            oxi_state_data.append(oxi_state)

        # add the new column to the dataframe
        dataframe = dataframe.assign(
            oxidation_state=oxi_state_data,
            element=elements,
        )
        # !!! There are multiple ways to do this, but I don't know which is best
        # dataframe["oxidation_state"] = pandas.Series(
        #     oxi_state_data, index=dataframe.index)

        # write output files/plots for the user to quickly reference
        cls._write_output_summary(directory, dataframe, extra_data)

        # return all of our results
        return dataframe, extra_data

    @staticmethod
    def _write_output_summary(
        directory: str,
        dataframe: DataFrame,
        extra_data: dict,
    ):
        """
        This prints a "simmate_summary.yaml" file with key output information.

        This method should not be called directly as it used within workup().
        """

        # write output of the dataframe
        summary_csv_filename = os.path.join(directory, "simmate_summary_bader.csv")
        dataframe.to_csv(summary_csv_filename)

        summary = {
            "notes": "view simmate_summary_bader.csv for more information",
            **extra_data,
        }

        summary_filename = os.path.join(directory, "simmate_summary.yaml")
        with open(summary_filename, "w") as file:
            content = yaml.dump(summary)
            file.write(content)



# -*- coding: utf-8 -*-

from enum import Enum

from simmate.materials_science.element_data import ALL_DATA, SUPPORTED_PROPERTIES


class Element(Enum):
    """
    The Element class is really just a dictionary of fixed data, but we perform some
    extra python "tricks" to help with performance and usage.

    WARNING: This class implements some advanced python topics, and you don't need
    to understand them to use this class as a beginner. Check out element_data.py
    (and the csv file) to see which data is available. Also look at this link for a
    quick tutorial:
        << TODO: INSERT LINK >>

    The main performance boost comes from making this a subclass of Enum, which
    means there are fixed number of options to choose from. Further, once one is
    chosen, it never changes. This means we really only need one object instance
    for each element -- and any time we call Element("Al"), it will share that instance
    and data. This is effectively a "singleton" and helps us save on memory and
    object initialization in a big way.

    The second thing we do is "lazy-loading" of an Element's attributes. This helps
    with how fast we can create Element objects. In 99% of cases, we really don't need
    all of the data for an element -- just the symbol. Therefore, data isn't loaded
    and attached to Element objects until they need to be. Once you call that attribute,
    we then link it to the object and it's there for good. This helps with performance
    too because once an attribute is accessed, it will likely be accessed again.
    """

    # These are the only possible Elements. Any other inputs for this class will
    # raise an error. Note, we also include a dummy element for when a site needs
    # to be specified without assigning an element to it: this is Unknown = "X"
    Hydrogen = "H"
    Helium = "He"
    Lithium = "Li"
    Beryllium = "Be"
    Boron = "B"
    Carbon = "C"
    Nitrogen = "N"
    Oxygen = "O"
    Fluorine = "F"
    Neon = "Ne"
    Sodium = "Na"
    Magnesium = "Mg"
    Aluminum = "Al"  # This is the correct spelling. Fight me about it.
    Silicon = "Si"
    Phosphorus = "P"
    Sulfur = "S"
    Chlorine = "Cl"
    Argon = "Ar"
    Potassium = "K"
    Calcium = "Ca"
    Scandium = "Sc"
    Titanium = "Ti"
    Vanadium = "V"
    Chromium = "Cr"
    Manganese = "Mn"
    Iron = "Fe"
    Cobalt = "Co"
    Nickel = "Ni"
    Copper = "Cu"
    Zinc = "Zn"
    Gallium = "Ga"
    Germanium = "Ge"
    Arsenic = "As"
    Selenium = "Se"
    Bromine = "Br"
    Krypton = "Kr"
    Rubidium = "Rb"
    Strontium = "Sr"
    Yttrium = "Y"
    Zirconium = "Zr"
    Niobium = "Nb"
    Molybdenum = "Mo"
    Technetium = "Tc"
    Ruthenium = "Ru"
    Rhodium = "Rh"
    Palladium = "Pd"
    Silver = "Ag"
    Cadmium = "Cd"
    Indium = "In"
    Tin = "Sn"
    Antimony = "Sb"
    Tellurium = "Te"
    Iodine = "I"
    Xenon = "Xe"
    Caesium = "Cs"
    Barium = "Ba"
    Lanthanum = "La"
    Cerium = "Ce"
    Praseodymium = "Pr"
    Neodymium = "Nd"
    Promethium = "Pm"
    Samarium = "Sm"
    Europium = "Eu"
    Gadolinium = "Gd"
    Terbium = "Tb"
    Dysprosium = "Dy"
    Holmium = "Ho"
    Erbium = "Er"
    Thulium = "Tm"
    Ytterbium = "Yb"
    Lutetium = "Lu"
    Hafnium = "Hf"
    Tantalum = "Ta"
    Tungsten = "W"
    Rhenium = "Re"
    Osmium = "Os"
    Iridium = "Ir"
    Platinum = "Pt"
    Gold = "Au"
    Mercury = "Hg"
    Thallium = "Tl"
    Lead = "Pb"
    Bismuth = "Bi"
    Polonium = "Po"
    Astatine = "At"
    Radon = "Rn"
    Francium = "Fr"
    Radium = "Ra"
    Actinium = "Ac"
    Thorium = "Th"
    Protactinium = "Pa"
    Uranium = "U"
    Neptunium = "Np"
    Plutonium = "Pu"
    Americium = "Am"
    Curium = "Cm"
    Berkelium = "Bk"
    Californium = "Cf"
    Einsteinium = "Es"
    Fermium = "Fm"
    Mendelevium = "Md"
    Nobelium = "No"
    Lawrencium = "Lr"
    Rutherfordium = "Rf"
    Dubnium = "Db"
    Seaborgium = "Sg"
    Bohrium = "Bh"
    Hassium = "Hs"
    Meitnerium = "Mt"
    Darmstadtium = "Ds"
    Roentgenium = "Rg"
    Copernicium = "Cn"
    Nihonium = "Nh"
    Flerovium = "Fl"
    Moscovium = "Mc"
    Livermorium = "Lv"
    Tennessine = "Ts"
    Oganesson = "Og"
    Unknown = "X"

    def __init__(self, symbol):

        self.symbol = symbol

        # OPTIMIZE: consider adding key attributes here for speed. The cost of
        # adding upfront is ___ns while adding later costs ___ns
        # I could have an enviornment variable so users can set this directly too.
        # That would be advanced usage, but may be useful to some.
        # For example...
        # self.atomic_number = ALL_DATA[symbol]["atomic_number"]

    def __getattr__(self, item):
        """
        This function is only called when "self.item" fails. It therefore
        takes longer than if you were to set the attribute in the init. However,
        more often than not, we don't need the attributes set prior. We do this
        to speed up the initialization of Element objects. This ends up being a
        problem later if the user is constantly trying to access an attribute that
        wasn't pre-set. As a solution to that, we set the attributes as they
        are accessed -- via the setattr() function. This means the first call takes
        longer and then it's fast after that. If I never access this attribute again,
        this is also wasted overhead.
        """

        # See if the requested attribute is actually one we support
        if item in SUPPORTED_PROPERTIES:

            # Grab our target data
            # NOTE: I am using data.get(item) instead of data[item] because I want
            # to return None when the data is available. This is slower but gives
            # our desired effect.
            value = ALL_DATA[self.symbol].get(item)

            # Set the attribute to the data so we can access it quickly in the future.
            setattr(self, item, value)

            # And now that everything is set up, return the target value to the user
            return value

        # if not, raise an error
        raise AttributeError(f"Element has no attribute {item}!")

    def __str__(self):
        """Just return the symbol if we want the Element as a String"""
        return self.symbol

    def __lt__(self, other):
        """
        Establishes how we evalutate "Element1 < Element2".

        This allows us to sort atomic species by electronegativity and then
        by symbol for equal electronegativities. Very useful for getting correct
        formulas. For example, FeO4PLi is automatically sorted into LiFePO4.
        """
        # NOTE: list sorting only uses less-than comparisons, so we don't need to
        # define a __gt__ method.

        # Not all elements have electronegativities, so we grab them first. If
        # an element doesn't have one, we set it to infinity. Otherwise, an
        # error will be thrown if we compare None with a float.
        x1 = self.electronegativity if self.electronegativity else float("inf")
        x2 = other.electronegativity if other.electronegativity else float("inf")

        # if the electronegativities are not equal, return whether the "other"
        # is less-than or not.
        if x1 != x2:
            return x1 < x2
        # If electronegativities are equal, we then sort alphabetically by symbol.
        else:
            return self.symbol < other.symbol

    @staticmethod
    def from_atomic_number(atomic_number):
        """Returns an element from an atomic number."""

        # iterate through each element until we find the one where numbers match
        for symbol, data in ALL_DATA.items():
            if data["atomic_number"] == atomic_number:
                return Element(symbol)

        # if we never find a match, raise an error
        raise ValueError(f"No element with atomic number {atomic_number}")

    @staticmethod
    def from_row_and_group(row_number, group_number):
        """
        Returns an element from a row and group number.
        Note that the 18 group number system is used, i.e., Noble gases are group 18.
        """

        # iterate through each element until we find the one where numbers match
        for symbol, data in ALL_DATA.items():
            if (
                data["row_number"] == row_number
                and data["group_number"] == group_number
            ):
                return Element(symbol)

        # if we never find a match, raise an error
        raise ValueError(f"No element with row {row_number} and group {group_number}")

    def all_data(self):
        """Returns the full dictionary of data available for the element"""
        return ALL_DATA[self.symbol]

    # OPTIMIZE: consider changing these properties to check group/row numbers instead
    # of comparing lists. For example, noble gases would check "group_number == 18".

    @property
    def is_noble_gas(self):
        """True if element is noble gas."""
        noble_gases = (2, 10, 18, 36, 54, 86, 118)
        return self.atomic_number in noble_gases

    @property
    def is_transition_metal(self):
        """True if element is a transition metal."""
        transition_metals = (
            list(range(21, 31))
            + [57]
            + list(range(72, 81))
            + [89]
            + list(range(104, 113))
        )
        return self.atomic_number in transition_metals

    @property
    def is_post_transition_metal(self):
        """True if element is a post-transition or poor metal."""
        post_transition_metals = (13, 31, 49, 50, 81, 82, 83)
        return self.atomic_number in post_transition_metals

    @property
    def is_metalloid(self):
        """True if element is a metalloid."""
        metalloids = (5, 14, 32, 33, 51, 52, 84)
        return self.atomic_number in metalloids

    @property
    def is_alkali(self):
        """True if element is an alkali metal."""
        alkali_metals = (3, 11, 19, 37, 55, 87)
        return self.atomic_number in alkali_metals

    @property
    def is_alkaline(self):
        """True if element is an alkaline earth metal"""
        alkali_earth_metals = (4, 12, 20, 38, 56, 88)
        return self.atomic_number in alkali_earth_metals

    @property
    def is_halogen(self):
        """True if element is a halogen."""
        halogens = (9, 17, 35, 53, 85)
        return self.atomic_number in halogens

    @property
    def is_chalcogen(self):
        """True if element is a chalcogen."""
        chalcogens = (8, 16, 34, 52, 84)
        return self.atomic_number in chalcogens

    @property
    def is_lanthanoid(self):
        """True if element is a lanthanoid."""
        return 56 < self.atomic_number < 72

    @property
    def is_actinoid(self):
        """True if element is a actinoid."""
        return 88 < self.atomic_number < 104

    @property
    def is_rare_earth_metal(self):
        """True if element is a rare earth metal."""
        return self.is_lanthanoid or self.is_actinoid

    @property
    def is_metal(self):
        """True if element is a metal."""
        # if any of these are true, it's a metal
        return (
            self.is_alkali
            or self.is_alkaline
            or self.is_post_transition_metal
            or self.is_transition_metal
            or self.is_lanthanoid
            or self.is_actinoid
        )

    @property
    def is_quadrupolar(self):
        """Checks if this element can be quadrupolar"""
        # If the element has nmr_quadrupole_moment data, it can be quadrupolar.
        # Otherwise it is not. We use bool() return true/false
        return bool(self.nmr_quadrupole_moment)



# -*- coding: utf-8 -*-

"""
This file is required!

You typically don't have anything extra other than what is shown below. All this
does is tell django where to find your code.
"""

from django.apps import AppConfig


class ExampleAppConfig(AppConfig):
    name = "example_app"
